# The world is a happy place

import pygame
from pygame.locals import *

import os
import math
import time
import re
import sys
import random
GAME_NAME = "Shackled Stones"
DEBUG_SCRIPTS = True

testCode = True
testCode = False

WARP_NSCROLL  = "north scroll"
WARP_SSCROLL  = "south scroll"
WARP_ESCROLL  = "east scroll"
WARP_WSCROLL  = "west scroll"
WARP_PIXELATE = "pixelate"
WARP_INSTANT  = "instant"
WARP_FADE     = "fade"

KEY_LEFT  = 'left'
KEY_RIGHT = 'right'
KEY_UP    = 'up'
KEY_DOWN  = 'down'

WHITE = pygame.Color('#ffffff')
BLACK = pygame.Color('#000000')
BLUE = pygame.Color('#0000ff')
RED = pygame.Color('#ff0000')

D_TEXT_OFFSET_X = 13
D_ANSWER_OFFSET_X = 25
D_TEXT_OFFSET_Y = 198

MENU_FONT = os.path.join('media', 'fortunaschwein.ttf')
TEXT_FONT = os.path.join('media', 'rm_typewriter_old.ttf')

ALPHA = 128


def trim(string):
	if string == None: return ''
	while len(string) > 0 and string[0] in ' \n\r\t':
		string = string[1:]
	while len(string) > 0 and string[-1] in ' \r\n\t':
		string = string[:-1]
	return string

def min(a, b):
	if a < b: return a
	return b
def max(a, b):
	if a < b: return b
	return a
def abs(a):
	if a < 0: return -a
	return a
	
def go_script_go(script_contents):
	script_contents = script_contents.replace('\\n', '\n') # -_-
	MapScript(ScriptIter(script_contents.split('\n'))).Exec()

def scriptPath(*path):
   p = os.path.join('data', 'scripts', *path)
   p = '%s.scr' % (p)
   return p

def portraitPath(*path):
   p = os.path.join('images', 'portraits', *path)
   p = '%s.png' % p
   return p

def uiImgPath(*path):
   p = os.path.join('images', 'ui', *path)
   p = '%s.png' % p
   return p

def render_text(string, color = BLACK):
   return _font.render(string, True, color)

_fontBucket = {}
def render_text_size(size, string, color = BLACK, fontPath = MENU_FONT):
   fontKey = '%s-%s' % (fontPath, str(size))
   f = _fontBucket.get(fontKey)
   
   if not f:
      f = pygame.font.Font(fontPath, size)
      _fontBucket[fontKey] = f
   return f.render(string, True, color)

def render_number(num, color=BLACK, sz=15):
   return render_text_size(sz, str(num), color, TEXT_FONT)

	
def make_list(size):
	return [None] * size

def make_table(width, height):
	cols = make_list(width)
	i = 0
	while i < width:
		cols[i] = make_list(height)
		i += 1
	return cols
	

def get_money():
	value = ActiveGame().GetVar('money_amount')
	if value == None:
		return 0
	else:
		return int(value)

def set_money(amount):
	ActiveGame().SetSavedVar('money_amount', int(amount))

def modify_money(amount):
	set_money(max(0, get_money() + amount))

def has_money(amount):
	return get_money() >= amount

def get_life():
	value = ActiveGame().GetVar('life_meter')
	if value == None:
		return 3
	else:
		return max(0, int(value))

def get_max_life():
	value = ActiveGame().GetVar('has_armor')
	if value == None or value == 0:
		return 10
	else:
		return 20
		
def set_life(amount):
	ActiveGame().SetTempVar('life_meter', min(10, amount))
def take_damage(amount):
	set_life(get_life() - amount)
	return get_life() <= 0
def heal_damage():
	set_life(get_life() + 1)

### .\source\aaaa.py

testSet = []

class Test:
   def __init__(self, name, fn):
      self._name = name
      self._fn = fn
   
   def __call__(self):
      print('\n%s\n------------------------------------------------------------' % self._name)
      self._fn()

def AddTest(name, fn):
   testSet.append(Test(name, fn))
   
def RunTests():
   for test in testSet:
      test()

### .\source\aScripted.py

class Scripted:
   def __init__(self, scriptIter):
      if not isinstance(scriptIter, ScriptIter):
         raise Exception("scriptIter must be an object of type ScriptIter")
      self._script = scriptIter
      self._fnTable = {}
      
      self._addFn('comment', self._noop)
      self._addFn('label', self._noop)
      self._addFn('jump', self._jump)
      self._addFn('check', self._checkVar)
      self._addFn('set', self._set)
      self._addFn('end', self._end)
      self._addFn('switch scene', do_switch_scene)
      self._addFn('buy', do_buy)
      self._addFn('save', self.save_game_foo)
      self._addFn('credits', self._credits)
      
   def _credits(self):
      ActiveGame().GetActiveGameScene().gotocredits = True
      
   def save_game_foo(self):
    
    ActiveGame().SetSavedVar('save_map', game_scene.name)
    ActiveGame().SetSavedVar('save_x', game_scene.player.x)
    ActiveGame().SetSavedVar('save_y', game_scene.player.y)
    ActiveGame().SaveToFile()
      
   def _parse(self):
      for line in self._script:
         if DEBUG_SCRIPTS:
            print(line)
         if Parser.IsCommand(line):
            (cmd, args) = Parser.Segment(line)
            if self._fnTable.get(cmd):
               c = self._call(cmd, args)
               
               if not c:
                  break
            else:
               print('Unrecognized command: %s' % cmd)
         else:
            self._parseInternal(line)

   def _parseInternal(self, line):
      print('NOT_IMPLEMENTED, dropping %s' % line)
   
   def _addFn(self, name, fn):
      self._fnTable[name] = fn
   
   def _call(self, name, args):
      if name in self._fnTable:
         return self._fnTable[name](*args)
      else:
         print('%s not registered' % name)

   # Move the script on if it's a multi-part deal.  Override this if you want
   # take special actions on script resume
   def Advance(self):
      self._parse()

   # function implementations
   # return indicates if script execution should continue (True) or stop until
   # the next call to Advance(False)
   def _checkVar(self, var, test, val, label, failLabel=None):
      sval = ActiveGame().GetVar(var)
      if test == 'eq':
         ret = (sval == val)
      elif test == 'lt':
         ret = (sval < val)
      elif test == 'lte.':
         ret = (sval <= val)
      elif test == 'gt':
         ret = (sval > val)
      elif test == 'gte':
         ret = (sval >= val)
      if ret:
         self._script.FindLabel(label)
      else:
         if failLabel:
            self._script.FindLabel(failLabel)
      return True
   
   def _set(self, var, val):
      ag = ActiveGame().SetSavedVar(var, val)
      return True

   def _jump(self, label):
      self._script.FindLabel(label)
      return True

   def _noop(self, *args):
      return True
   
   def _end(self):
      return False

### .\source\Dialog.py

D_NORMAL   = 0
D_SCAN     = 1
D_PAUSE    = 2
D_QUESTION = 3
D_CHECKVAR = 4 
D_END      = 5

class Dialog(Scripted):
   # script -- ScriptIter
   def __init__(self, script):
      Scripted.__init__(self, script)
      self._profile = None
      self._state = D_NORMAL
      
      # What we should be displaying
      self._buffer = ''
      # When we're in question mode
      self._question = None
      # when we have a question what are the choices?
      self._choices = []
      
      self._addFn('profile', self._setProfile)
      self._addFn('pause', self._pause)
      self._addFn('question', self._beginQuestion)
      self._addFn('choice', self._addChoice)
      self._addFn('/question', self._poseQuestion)
      self._addFn('save', self._saveGame)
      self._addFn('credits', self._credits)
      
      # perform the initial parse (fill the buffer)
      self.Advance()
   
   # Get the path to the current profile
   def Profile(self):
      return self._profile
   
   def _credits(self):
      ActiveGame().GetActiveGameScene().gotocredits = True
   
   # Find out what mode the dialog is in
   def State(self):
      return self._state
   
   # get the next bit of stuff to display
   def Advance(self):
      # do not allow resuming if the dialog is finished
      if self.State() == D_END:
         return
      
      self._buffer = ''
      Scripted.Advance(self)
   
   # What we should be displaying if we're in "talk" mode (D_NORMAL)
   def Text(self):
      return trim(self._buffer)
   
   # what choices are available
   def Choices(self):
      if self.State() != D_QUESTION:
         print("ERR: Not in Question mode")
      options = []
      for a in self._choices:
         options.append(a.Text())
      return options
   
   # Answer a question
   # resp - which choice the user went with
   def Answer(self, resp):
      c = self._choices[resp]
      self._choices = []
      self._state = D_NORMAL
      self._script.FindLabel(c.Label())
   
   def _parseInternal(self, line):
      if len(line) == 0:
         return
         
      if line == '\\n':
         line = '$nl$'
      name = ActiveGame().GetVar('name')
      line = line.replace('%Name%', name).replace('%NAME%', name)
      self._buffer += line + '\n'

   # function implementations
   # return indicates if script execution should continue (True) or stop until
   # the next Advance (False)
   def _beginQuestion(self, text):
      self._state = D_QUESTION
      self._choices = []
      self._buffer = text
      return True
   
   def _addChoice(self, label, text):
      self._choices.append(Choice(text, label))
      return True
   
   def _saveGame(self):
      game_scene = ActiveGame().GetActiveGameScene()
      ActiveGame().SetSavedVar('save_map', game_scene.name)
      ActiveGame().SetSavedVar('save_x', game_scene.player.x)
      ActiveGame().SetSavedVar('save_y', game_scene.player.y)
      ActiveGame().SaveToFile()
      return True
   
   def _poseQuestion(self):
      return False
   
   def _setProfile(self, file):
      self._profile = file
      return True
   
   def _end(self):
      self._state = D_END
      return Scripted._end(self)
   
   def _pause(self):
      return False

class Choice:
   def __init__(self, txt, label):
      self._text = txt
      self._label = label
   def __str__(self):
      return '[%s] %s' % (self.Label(), self.Text())
   
   def Text(self):
      return self._text
   def Label(self):
      return self._label
   
###############################################################################
# Testing Code

def testDialog():
   si = Parser.LoadFile(scriptPath('test'))
   d = Dialog(si)
   
   print('----------------------')
   print('current dialog text:')
   print("'%s'" % d.Text())
   print('----------------------')
   
   d.Advance()
   print('----------------------')
   print('current dialog text:')
   print("'%s'" % d.Text())
   print('----------------------')
   
AddTest('testDialog', testDialog)

### .\source\GameContext.py

"""
common variables:

name - player name

The following indicate which items the player has. They are either 1 or None

item_sabre
item_hammer
item_drill
item_hook
item_cannon
item_cannon_fire
item_cannon_ice
item_cannon_multi
item_compass
item_shovel

The following holds one of the above to indicate what each button has assigned
equipped_a
equipped_b
equipped_x
equipped_y

The following will be used to track which stones a play has acquired
stone_water
stone_light
stone_life
stone_fire
stone_dark
stone_death

The following will be used to track which temples have been completed
A temple is complete when the stone is replaced on the altar
finish_water
finish_light
finish_dark
finish_fire
finish_life
finish_death
"""
class GameInstance:
	def __init__(self, slot):
		self.slot = slot
		self.values = self.parse(slot)
		self.temp_vars = {}
		self.active_game_scene = None
	
	def GetVar(self, varName):
		value = self.temp_vars.get(varName)
		if value == None:
			return self.values.get(varName)
		return value
	
	def SetActiveGameScene(self, game_scene):
		self.active_game_scene = game_scene
	def GetActiveGameScene(self):
		return self.active_game_scene
	
	def SetSavedVar(self, name, value):
		self.values[name] = value
	def SetTempVar(self, name, value):
		self.temp_vars[name] = value
   
	def parse(self, slot=None):
		if slot == None:
			slot = self.slot
		values = {}
		c = open('saves' + os.sep + 'slot' + str(slot) + '.txt', 'rt')
		lines = trim(c.read()).split('\n')
		c.close()
		for line in lines:
			parts = trim(line).split(':')
			if len(parts) > 1:
				name = trim(parts[0])
				value = trim(':'.join(parts[1:]))
				if len(name) > 1 and name[0] in '$#':
					if name[0] == '#':
						value = int(value)
					name = name[1:]
					values[name] = value
		return values
	
	def SaveToFile(self):
		output = []
		for key in self.values.keys():
			value = self.values[key]
			if str(value) == value:
				name = '$' + key
			else:
				name = '#' + key
			output.append(name + ':' + str(value))
		c = open('saves' + os.sep + 'slot' + str(self.slot) + '.txt', 'wt')
		c.write('\r\n'.join(output))
		c.close()

class GameContext:
	
	def __init__(self):
		for i in (1, 2, 3):
			file = 'saves' + os.sep + 'slot' + str(i) + '.txt'
			if not os.path.exists(file):
				c = open(file, 'wt')
				c.write(' ')
				c.close()
		self.slots = [GameInstance(1), GameInstance(2), GameInstance(3)]
		self.active_game = None

	def SetActiveGame(self, slot_num):
		self.active_game = self.slots[slot_num - 1]
		
	def GetActiveGame(self):
		return self.active_game
	
	def SetPlayerName(self, slot_num, name):
		self.slots[slot_num - 1].SetSavedVar('name', name)
	
	def GetPlayerName(self, slot_num):
		return self.slots[slot_num - 1].GetVar('name')
	def GetStones(self, slot_num):
		r = []
		s = self.slots[slot_num - 1]
		for st in ['water', 'light', 'dark', 'fire', 'life', 'death']:
			if s.GetVar('stone_%s' % st):
				r.append(st)
		return r
		
	
	def DeletePlayer(self, slot_num):
		game = self.slots[slot_num - 1]
		
		#shut up
		game.temp_vars = {}
		game.values = {}
		
		game.SaveToFile()
	
	def CopyPlayer(self, from_slot, to_slot):
		if from_slot != to_slot:
			from_game = self.slots[from_slot - 1]
			to_game = self.slots[to_slot - 1]
			
			to_game.values = from_game.values
			to_game.SaveToFile()
			
			# this will ensure that the dictionary instance is different from from_game
			self.slots[to_slot - 1] = GameInstance(to_slot)
			


### .\source\GamePlay\CompassRenderer.py

class CompassRenderer:
	
	def __init__(self):
		self.counter = 0
		self.on_temple = False
		
	
	def Render(self, screen, game_scene):
		compass_image = get_image('ui/compass')
		self.counter += 1
		game_scene = ActiveGame().GetActiveGameScene()
		if ActiveGame().GetVar('is_compass_active') == 1:
			screen.blit(compass_image, (10, 200))
			spin_angle = self.counter * 3.14159 * 2 / 30
			if game_scene != None and game_scene.level.dungeon == 'light':
				angle = spin_angle
			elif game_scene == None or game_scene.name != 'world_W':
				angle = 3.14159 / 2
			else:
				tile = game_scene.level.ids['temple']
				x = (tile.x << 4) + 8
				y = (tile.y << 4) + 8
				player_x = game_scene.player.x 
				player_y = game_scene.player.y
				
				dx = x - player_x
				dy = y - player_y
				
				if abs(dx) < 5 and abs(dy) < 5:
					angle = spin_angle
				elif dx == 0:
					if dy < 0:
						angle = 3.14159 / 2
					else:
						angle = -3.14159 / 2
				elif dy == 0:
					if dx < 0:
						angle = 0
					else:
						angle = 3.14159
				else:
					angle = math.atan(dy / (0.0+ dx))
					if dx > 0:
						angle += 3.14159
			
			angle -= 3.14159 / 2
			n = (math.cos(angle - 3.14159 / 2), math.sin(angle - 3.14159 / 2))
			s = (math.cos(angle + 3.14159 / 2), math.sin(angle + 3.14159 / 2))
			e = (math.cos(angle) / 3, math.sin(angle) / 3)
			w = (math.cos(angle + 3.14159) / 3, math.sin(angle + 3.14159) / 3)
			
			
			n = (int(n[0] * 20) + 40, int(n[1] * 20) + 230)
			s = (int(s[0] * 20) + 40, int(s[1] * 20) + 230)
			e = (int(e[0] * 20) + 40, int(e[1] * 20) + 230)
			w = (int(w[0] * 20) + 40, int(w[1] * 20) + 230)
			
			pygame.draw.polygon(screen, (180, 0, 0), [n, e, w])
			pygame.draw.polygon(screen, (255, 255, 255), [s, e, w])
			pygame.draw.polygon(screen, (0,0,0), [n, e, s, w], 2)
			

### .\source\GamePlay\CutScene.py

class CutSceneEvent:
	def __init__(self, command):
		parts = trim(command).split(' ')
		args = parts[1:]
		name = parts[0].lower()
		self.expiration = -1
		self.key_pressed = ''
		self.render_offset = 0
		self.direction = 'xxxxxxxxxx'
		self.first_time_through = True
		self.instant = False
		if name == 'pause':
			self.do = self.pause
			self.expiration = int(args[0])
		elif name == 'clearbg':
			self.do = self.do_clearbg
			self.instant = True
		elif name == 'setbg':
			self.do = self.do_setbg
			self.instant = True
			self.image = args[0]
		elif name == 'flash':
			self.do = self.do_flash
			self.expiration = int(args[0])
			self.full_length = self.expiration
		elif name == 'save':
			self.do = self.save
		elif name == 'script':
			self.do = self.do_script
			self.script = ' '.join(args)
			self.instant = True
		elif name == 'input':
			self.do = self.do_input
			self.button = args[0]
			self.key_pressed = self.button
			self.expiration = int(args[1])
		elif name == 'dialog':
			self.do = self.show_dialog
			self.dialog_script = args[0]
			self.expiration = -1
		elif name == 'playsound':
			self.do = self.do_play_sound
			self.sound = args[0]
			self.instant = True
		elif name == 'shakescreen':
			self.expiration = int(args[0])
			self.play_sound_on = self.expiration
			self.do = self.shake_screen
		elif name == 'relighttorches':
			self.instant = True
			self.do = self.do_relighttorches
		elif name == 'turnlightswitchesoff':
			self.instant = True
			self.do = self.do_turnlightswitchesoff
		elif name == 'sprite':
			if args[0].lower() == 'delete':
				self.sprite_id =args[1]
				self.do = self.deletesprite
				self.instant=True
			if args[0].lower() == 'setstate':
				self.sprite_id = args[1]
				self.sprite_state = args[2]
				self.do = self.setspritestate
				self.instant = True
			elif args[0].lower() == 'create':
				self.sprite_name = args[1]
				self.sprite_id = args[2]
				self.x = int(args[3])
				self.y = int(args[4])
				self.layer = args[5]
				self.state = args[6]
				self.do = self.createsprite
				self.direction = args[7]
				self.instant = True
			elif args[0].lower() == 'setdirection':
				self.sprite_id = args[1]
				self.direction = args[2]
				#print self.direction
				self.do = self.setspritedirection
				self.instant = True
			elif args[0].lower() == 'setxy':
				self.sprite_id = args[1]
				self.do = self.setspriteloc
				self.x = int(args[2])
				self.y = int(args[3])
				self.instant = args[4].lower() == 'instant'
				self.expiration = -1
				if not self.instant:
					self.expiration = int(args[4])
					self.total = self.expiration
	
	def do_turnlightswitchesoff(self, game_scene):
		game_scene.turnlightswitchesoff()
	
	def do_script(self, game_scene):
		go_script_go(self.script)
	
	def do_setbg(self, game_scene):
		game_scene.bg = get_image(self.image)
	
	def do_clearbg(self, game_scene):
		game_scene.bg = None
	
	def do_flash(self, game_scene):
		antiprogress = min(1.0, max(0.0, (0.0 + self.expiration) / self.full_length))
		
		antiprogress = antiprogress * 2 - 1
		
		antiprogress = 1.0 - abs(antiprogress)
		#antiprogress = 1.0 - antiprogress
		game_scene.flash_amount = abs(antiprogress)
	
	def do_play_sound(self, game_scene):
		play_sound(self.sound)
	
	def do_relighttorches(self, game_scene):
		game_scene.torch_puzzle_relight()
	
	def save(self, game_scene):
		ActiveGame().SetSavedVar('save_map', game_scene.name)
		ActiveGame().SetSavedVar('save_x', game_scene.player.x)
		ActiveGame().SetSavedVar('save_y', game_scene.player.y)
		ActiveGame().SaveToFile()
	
	def shake_screen(self, game_scene):
		if self.expiration == self.play_sound_on:
			pass
			#print("TODO: play shaking sound")
		self.render_offset = (-2,2)[(self.expiration & 1) == 0]
		
	def is_key_pressed(self, key):
		return self.key_pressed == key
		
	def pause(self, game_scene):
		pass
	
	def do_input(self, game_scene):
		pass
	
	def setspritedirection(self, game_scene):
		sprite = game_scene.get_sprite_by_id(self.sprite_id)
		sprite.direction = self.direction
	
	def createsprite(self, game_scene):
		sprite = create_sprite(self.sprite_name, self.sprite_id)
		sprite.x = (self.x << 4) + 8
		sprite.y = (self.y << 4) + 8
		sprite.state = self.state
		sprite.layer = self.layer
		sprite.direction = self.direction
		game_scene.add_sprite(sprite)
	
	def deletesprite(self, game_scene):
		sprite = game_scene.get_sprite_by_id(self.sprite_id)
		sprite.expired=True
	
	def show_dialog(self, game_scene):
		game_scene.next = DialogScene(
			Dialog(Parser.LoadFile(scriptPath(self.dialog_script))), game_scene)
	
	def setspritestate(self, game_scene):
		sprite = game_scene.get_sprite_by_id(self.sprite_id)
		sprite.state = self.sprite_state
		
	def setspriteloc(self, game_scene):
		sprite = game_scene.get_sprite_by_id(self.sprite_id)
		if not sprite:
			return
		if self.first_time_through:
			self.source_x = sprite.x
			self.source_y = sprite.y
		target_x = (self.x << 4) + 8
		target_y = (self.y << 4) + 8
		if self.instant:
			sprite.x = target_x
			sprite.y = target_y
		else:
			progress = 1 - ((0.0 + self.expiration) / self.total)
			anti_progress = 1 - progress
			sprite.x = target_x * progress + anti_progress * self.source_x
			sprite.y = target_y * progress + anti_progress * self.source_y
			
		self.first_time_through = False
			
	
	
		
		

class CutScene:
	
	def __init__(self, command_script, name):
		self.name = name
		lines = trim(command_script).split('\n')
		commands = []
		for line in lines:
			command_line = trim(line)
			if len(command_line) > 0:
				commands.append(CutSceneEvent(command_line))
		self.commands = commands
		self.event_queue = []
	
	def render_offset(self):
		if len(self.commands) > 0:
			return self.commands[0].render_offset
	
	def is_done(self):
		return len(self.commands) == 0
	
	def is_key_pressed(self, key):
		return len(self.commands) > 0 and self.commands[0].is_key_pressed(key)
	
	def get_input_events(self):
		events = self.event_queue
		self.event_queue = []
		return events
	
	def do(self, game_scene):
		game_scene.flash_amount = 0
		while not self.is_done():
			command = self.commands[0]
			command.do(game_scene)
			command.expiration -= 1
			do_again = command.instant
			
			if command.expiration < 0:
				if command.key_pressed != '':
					self.event_queue.append(InputEvent(command.key_pressed, False))
				self.commands = self.commands[1:]
				if len(self.commands) > 0 and self.commands[0].key_pressed != '':
					self.event_queue.append(InputEvent(self.commands[0].key_pressed, True))
			if not do_again:
				return
				



### .\source\GamePlay\DeathCircle.py


class DeathCircle:
	def __init__(self, x, y, radius, duration, type):
		self.x = x
		self.y = y
		self.r = radius
		self.time_left = duration
		self.type = type
		self.expired = False
	
	def update(self):
		self.time_left -= 1
		if self.time_left < 0:
			self.expired = True
	
	def touches_sprite(self, sprite):
		if sprite.flash_counter < 0:
			dx = sprite.x - self.x
			dy = sprite.y - self.y
			max_distance = self.r + sprite.r
			if dx ** 2 + dy ** 2 < max_distance ** 2:
				return True
		return False


### .\source\GamePlay\GamePlayScene.py


class GamePlayScene:
	
	def __init__(self, level_name, startX, startY):
		global _new_enemies
		_new_enemies = []
		ActiveGame().SetActiveGameScene(self)
		print(level_name)
		self.render_counter = 0
		self.next = self
		self.player = Player()
		self.player.x = startX
		self.player.y = startY
		self.flash_amount = 0
		self.level = Level(level_name)
		self.death_circles = []
		self.disable_save = False
		self.name = level_name
		self.bg = None
		self.gotocredits = False
		self.player_invisible = False
		self.sprites = []
		self.overlayRenderer = OverlayRenderer()
		self.grapple = None
		self.cutscene = get_cutscene_for_map(level_name)
		self.last_torch_pressed = None
		self.lever_a_pressed = False
		self.lever_b_pressed = False
		self.puz_flag = False
		self.temp_screen = None
		
		self.prevTile = None
		self.firstTimeOnTile = True
		on_load_script = trim(self.level.on_load)
		self.inventory = Inventory()
		if len(on_load_script) > 0:
			go_script_go(on_load_script)
		
		if self.level.dungeon == 'light':
			mirrors = self.light_puzzle_get_mirror_states()
			setvar = ActiveGame().SetSavedVar
			setvar('mirror_door_open', 'None')
			if str(ActiveGame().GetVar('light_puzzle_on')) == '1':
				if mirrors['A'] == 'mirror1':
					if mirrors['B'] == 'mirror2' and mirrors['C'] == 'mirror3' and mirrors['D'] == 'mirror1' and mirrors['E'] == 'mirror3' and mirrors['F'] == 'mirror1':
						setvar('mirror_door_open', 'Blue')
				elif mirrors['A'] == 'mirror2':
					if mirrors['G'] == 'mirror1' and mirrors['H'] == 'mirror3':
						if mirrors['I'] == 'mirror4' and mirrors['L'] == 'mirror1' and mirrors['M'] == 'mirror2':
							setvar('mirror_door_open', 'Yellow')
						elif mirrors['I'] == 'mirror1' and mirrors['J'] == 'mirror4' and mirrors['K'] == 'mirror2':
							setvar('mirror_door_open', 'Red')
		
		self.torch_puz = level_name == 'dark_swamp'
		self.light_puz = level_name == 'light_puzzle1_f1'
		self.desert_puz = level_name == 'world_W'
		self.lever_puz = level_name == 'light_south_southroom'
		
		if self.torch_puz:
			self.swamp_opened = ActiveGame().GetVar('swamp_opened') != None
			
			if self.swamp_opened:
				self.open_dark_temple()
		
		self.initialize_enemies()
		play_music(self.level.music)
	
	def open_dark_temple(self):
		ids = self.level.ids
		go_script_go('\n'.join([
			'[set tile][1][doodad][d9]',
			'[set tile][2][doodad][d7]',
			'[set tile][3][doodad][d10]',
			'[set tile][4][doodad][d5]',
			'[set tile][5][doodad][94]',
			'[set tile][6][doodad][d4]',
			'[set tile][7][doodad][d12]',
			'[set tile][8][doodad][d164]',
			'[set tile][9][doodad][d11]',
			'[set tile][5][excessive][d41]'
			]))
		ActiveGame().SetSavedVar('swamp_opened', '1')
	
	def initialize_enemies(self):
		for enemy in self.level.enemies:
			data = trim(enemy)
			if len(data) > 0:
				parts = data.split('|')
				kind = parts[0]
				x = int(parts[2])
				y = int(parts[3])
				layer = parts[1]
				sprite = create_sprite(kind)
				sprite.x = (x << 4) + 8
				sprite.y = (y << 4) + 8
				sprite.layer = layer
				self.sprites.append(sprite)
				
	def place_player(self, layer, x, y):
		self.player.layer = layer
		self.player.x = (x << 4) + 8
		self.player.y = (y << 4) + 8
		self.level.synch_stand_key(layer, self.player.x >> 4, self.player.y >> 4)
	
	def is_key_pressed(self, key):
		if self.cutscene != None and not self.cutscene.is_done() and self.cutscene.name != 'asynch':
			return self.cutscene.is_key_pressed(key)
		return is_pressed(key)
			
	def add_sprite(self, sprite):
		self.sprites.append(sprite)
	
	def get_sprite_by_id(self, sprite_id):
		for sprite in self.sprites:
			if sprite.id == sprite_id:
				return sprite
		return None
	
	def place_death_circle(self, type, x, y, radius, duration):
		self.death_circles.append(DeathCircle(x, y, radius, duration, type))
		
		tile_x = x >> 4
		tile_y = y >> 4
		layer = self.level.layers[self.player.layer]
		if layer.contains_stuff and tile_x >= 0 and tile_y >= 0 and tile_x < layer.width and tile_y < layer.height:
			tile = layer.tiles[tile_x][tile_y]
			if type == 'sword':
				tile.ChopTheBushes()
			elif type == 'hammer':
				tile.SmashBoulders()
			elif type == 'drill':
				tile.DrillThrough()
			
	def ProcessInput(self, events):
	
		if self.cutscene != None and not self.cutscene.is_done() and self.cutscene.name != 'asynch':
			events = self.cutscene.get_input_events()
		
		actions = {
			'item_sabre' : False,
			'item_hammer' : False,
			'item_drill' : False,
			'item_hook' : False,
			'item_cannon' : False,
			'item_cannon_fire' : False,
			'item_cannon_ice' : False,
			'item_cannon_multi' : False,
			'item_compass' : False,
			'item_shovel' : False
		}
		
		items_pressed = []
		
		for event in events:
			if event.down and event.key == 'B':
				actions[self.inventory.EquippedB()] = True
			elif event.down and event.key == 'A':
				actions[self.inventory.EquippedA()] = True
			elif event.down and event.key == 'Y':
				actions[self.inventory.EquippedY()] = True
			elif event.down and event.key == 'X':
				actions[self.inventory.EquippedX()] = True
			elif event.down and event.key == 'start':
				self.next = InventoryScene(self)
		
		if self.player.state == 'walking':
			if actions['item_sabre']:
				self.player.Stab()
			elif actions['item_cannon']:
				self.player.Shoot('basic', self)
			elif actions['item_cannon_fire']:
				self.player.Shoot('fire', self)
			elif actions['item_cannon_ice']:
				self.player.Shoot('ice', self)
			elif actions['item_cannon_multi']:
				self.player.Shoot('multi', self)
			elif actions['item_shovel']:
				self.player.Dig()
			elif actions['item_drill']:
				self.player.Drill()
			elif actions['item_hammer']:
				self.player.Hammer()
			elif actions['item_hook']:
				self.player.Grapple()
			
		
		if actions['item_compass']:
			compass_active = ActiveGame().GetVar('is_compass_active')
			if compass_active == None: compass_active = 0
			else: compass_active = int(compass_active)
			
			if compass_active == 1:
				ActiveGame().SetTempVar('is_compass_active', 0)
			else:
				ActiveGame().SetTempVar('is_compass_active', 1)
		v = 3
		vx = 0
		vy = 0
		
		if self.player.state == 'walking':
			if self.is_key_pressed('left'):
				self.player.direction = 'left'
				vx = -v
			elif self.is_key_pressed('right'):
				self.player.direction = 'right'
				vx = v
			if self.is_key_pressed('up'):
				self.player.direction = 'up'
				vy = -v
			elif self.is_key_pressed('down'):
				self.player.direction = 'down'
				vy = v
		
		self.player.walking = False
		if vx != 0 or vy != 0:
			self.player.walking = True
		
		self.do_sprite_move(self.player, vx, vy, False)
		
	def Update(self, game_counter):
		self.level.update_tile_standing_on(self.player.layer, self.player.x, self.player.y)
		
		if self.prevTile != self.level.playerStandingOn:
			self.prevTile = self.level.playerStandingOn
			self.firstTimeOnTile = True
		else:
			self.firstTimeOnTile = False
		if self.torch_puz and not self.swamp_opened:
			self.torch_puzzle_update()
		elif self.lever_puz:
			self.light_lever_puzzle()
		if self.desert_puz:
			self.desert_puzzle_update()
		
		if self.level.dungeon != None and len(self.level.locked_doors) > 0:
			tile_x = self.player.x >> 4
			tile_y = self.player.y >> 4
			door = self.level.locked_doors.get(str(tile_x) + '_' + str(tile_y))
			if door != None:
				x = door[0]
				y = door[1]
				color = door[2]
				if GetKeyRegistry().UseKey(self.level.dungeon, color, self.level.name, x, y):
					self.level.RemoveLockedDoor(x, y)
		
		if self.gotocredits:
			self.next = CreditsScene()
		
		if self.cutscene != None and not self.cutscene.is_done():
			self.cutscene.do(self)
			if self.cutscene != None and self.cutscene.is_done():
				self.cutscene = None
		
		for sprite in self.get_sprites():
			sprite.Update()
			if sprite.dx != 0 or sprite.dy != 0:
				if self.do_sprite_move(sprite, sprite.dx, sprite.dy, sprite.flying):
					if sprite.explode_on_impact:
						sprite.expired = True
						if sprite.kind == 'grapple':
							if not self.grapple.grappled and not self.try_grapple_to(self.player):
								self.grapple = None
							else:
								self.grapple.expired = False
		
		if self.grapple != None:
			if self.grapple.grappled:
				self.grapple.grapple_counter += 1
				progress = (0.0 + self.grapple.grapple_counter) / self.grapple.grapple_duration
				antiprogress = 1 - progress
				self.player.x = int(self.grapple.x * progress + self.grapple.grapple_start_x * antiprogress)
				self.player.y = int(self.grapple.y * progress + self.grapple.grapple_start_y * antiprogress)
				if self.grapple.grapple_counter >= self.grapple.grapple_duration:
					self.grapple.expired = True
					self.grapple = None
			elif self.grapple.expired:
				self.grapple = None
			
		
		dcs = []
		for dc in self.death_circles:
			dc.update()
			if not dc.expired:
				dcs.append(dc)
		self.death_circles = dcs
		
		for sprite in self.sprites:
			if sprite.is_goody:
				dx = sprite.x - self.player.x
				dy = sprite.y - self.player.y
				if dx ** 2 + dy ** 2 < 16 ** 2:
					sprite.expired = True
					if sprite.name == 'money':
						modify_money(1)
					else:
						heal_damage()
						
		
		enemy_count = self.get_enemy_count()
		self.gc_sprites()
		if enemy_count > 0 and self.get_enemy_count() == 0:
			enemy_kill_script = self.level.on_enemies_killed
			if len(enemy_kill_script) > 0:
				go_script_go(enemy_kill_script)

	def try_grapple_to(self, player):
		if self.grapple:
			x = self.grapple.x
			y = self.grapple.y
			d = self.grapple.direction
			if d == 'left':
				x -= 8
			elif d == 'right':
				x += 8
			elif d == 'down':
				y += 8
			else:
				y -= 8
			x = x >> 4
			y = y >> 4
			layer = self.level.layers[self.grapple.layer]
			if x >= 0 and x < layer.width and y >= 0 and y < layer.height:
				tile = layer.tiles[x][y]
				if tile.is_grappleable:
					self.grapple.grappled = True
					return True
		return False

	def light_lever_puzzle(self):
		getvar = ActiveGame().GetVar
		
		if str(getvar('light_timed_lever_A')) == '1' and str(getvar('light_timed_lever_B')) == '1' and not self.puz_flag:
			go_script_go('[set][light_timed_lever_solved][1]')
			go_script_go('[remove tile][closed_door][baseadorn]')
			self.puz_flag = True
			self.cutscene = None
			
		if self.firstTimeOnTile:
			if not str(getvar('light_timed_lever_solved')) == '1':
				if str(getvar('light_timed_lever_A')) == '1' and not self.lever_a_pressed:
					self.lever_a_pressed = True
					if self.lever_b_pressed:
						go_script_go('[set][light_timed_lever_solved][1]')
						self.cutscene = None
					else:
						script = '[set][light_timed_lever_A][0]'
						self.cutscene = CutScene(('pause 15\nplaysound tick\n' * 8) + 'turnlightswitchesoff', 'asynch')
				if str(getvar('light_timed_lever_B')) == '1' and not self.lever_b_pressed:
					self.lever_b_pressed = True
					if self.lever_a_pressed:
						go_script_go('[set][light_timed_lever_solved][1]')
						self.cutscene = None
					else:
						script = '[set][light_timed_lever_B][0]'
						self.cutscene = CutScene(('pause 15\nplaysound tick\n' * 8) + 'turnlightswitchesoff', 'asynch')
				
				if str(getvar('light_timed_lever_A')) == '1' and str(getvar('light_timed_lever_B')) == '1':
					self.turnlightswitchesoff()
					go_script_go('[set][light_timed_lever_solved][1]')
	
	def turnlightswitchesoff(self):
		self.lever_a_pressed = False
		self.lever_b_pressed = False
		go_script_go('[set tile][switch_left][baseadorn][57]')
		go_script_go('[set tile][switch_right][baseadorn][57]')
		go_script_go('[set][light_timed_lever_A][0]')
		go_script_go('[set][light_timed_lever_B][0]')
		self.cutscene = None
		
	def desert_puzzle_update(self):
		opened = str(ActiveGame().GetVar('light_temple_opened')) == '1'
		if not opened and self.player.state == 'shovelling':
			target = self.level.ids['temple']
			x = self.player.x >> 4
			y = self.player.y >> 4
			if target.x == x and target.y == y:
				ActiveGame().SetSavedVar('light_temple_opened', '1')
				self.cutscene = CutScene('pause 5\nplaysound itemget\nscript [set tile][entrance][base][171]', 'open_light_temple')
			
		
		
	def get_enemy_count(self):
		count = 0
		for sprite in self.sprites:
			if sprite.is_enemy: count += 1
		return count
		
	def Render(self, screen):
		
		if self.bg != None:
			screen.blit(self.bg, (0,0))
		else:
			
			offset = self.get_camera_offset()
			
			
			if self.cutscene != None and not self.cutscene.is_done():
				r_offset = self.cutscene.render_offset()
				offset = (offset[0] + r_offset, offset[1])
			
			self.level.Render('Stairs', screen, offset[0], offset[1], self.render_counter)
			for layerName in 'A B C D E F Stairs'.split(' '):
				if layerName != 'Stairs':
					self.level.Render(layerName, screen, offset[0], offset[1], self.render_counter)
				
				for sprite in self.get_renderable_sprites(layerName):
					img = sprite.CurrentImage(self.render_counter)
					if img != None:
						coords = sprite.DrawingCoords()
						screen.blit(img, (coords[0] + offset[0], coords[1] + offset[1]))
					if sprite == self.grapple:
						x = sprite.x
						y = sprite.y
						if sprite.direction == 'left':
							x += 14
						elif sprite.direction == 'right':
							x -= 14
						elif sprite.direction == 'up':
							y += 14
						else:
							y -= 14
						pygame.draw.line(screen, (255, 200,40), (sprite.x, sprite.y), (self.player.x, self.player.y))
					if sprite.is_enemy and sprite.frozen and img != None:
						pygame.draw.rect(screen, (100, 100, 255), Rect(coords[0] + offset[0]- 2, coords[1] + offset[1]-2, sprite.r * 2  +4, sprite.r * 2+4), 1)
			
			
			
			if self.light_puz:
				self.render_light_puzzle(screen, offset)
			
		if self.cutscene != None:
			self.make_white(screen, self.flash_amount)
		else:
			self.flash_amount = 0
		
		
		self.render_counter += 1
		
		if self.overlayRenderer != None:
			self.overlayRenderer.Render(screen)
	
	def make_white(self, screen, amount):
		if amount == 0: return
		if self.temp_screen == None:
			self.temp_screen = pygame.Surface((screen.get_width(), screen.get_height()))
		value = min(255, max(0, 255 * amount))
		self.temp_screen.fill((255,255,255))
		self.temp_screen.set_alpha(value)
		screen.blit(self.temp_screen, (0,0))
		
	def torch_puzzle_update(self):
		ids = self.level.ids
		activation = 'ABCDEFGHI'
		x = self.player.x >> 4
		y = self.player.y >> 4
		current = None
		
		last = self.last_torch_pressed
		for active_tile in activation:
			tile = ids[active_tile]
			if tile.x == x and tile.y == y:
				current = active_tile
				break
		
		name = None
		if self.cutscene != None: name = self.cutscene.name
		
		if self.firstTimeOnTile and current != None and current != last and self.cutscene == None:
			if last == 'H' and current == 'I':
				self.cutscene = get_cutscene('torch_win')
				self.open_dark_temple()
			elif last == None and current == 'A':
				self.last_torch_pressed = current
				play_sound('fwuf')
				go_script_go('[remove tile][t' + current + '][doodad]')
			elif last != None and activation.find(last) == activation.find(current) - 1:
				self.last_torch_pressed = current
				play_sound('fwuf')
				go_script_go('[remove tile][t' + current + '][doodad]')
			else:
				play_sound('fwuf')
				go_script_go('[remove tile][t' + current + '][doodad]')
				current = None
				last = None
				self.last_torch_pressed = None
				self.cutscene = get_cutscene('torch_fail')
	
	def torch_puzzle_relight(self):
		for tile in 'ABCDEFGHI':
			go_script_go('[set tile][t' + tile + '][doodad][torch3]')
		play_sound('bad')
			
	
	def render_light_puzzle(self, screen, offset):
		get_var = ActiveGame().GetVar
		ids = self.level.ids
		
		mirror_states = self.light_puzzle_get_mirror_states()
		mirror_images = {
			'mirror1' : get_image('tiles/mirrors/topright'),
			'mirror2' : get_image('tiles/mirrors/topleft'),
			'mirror3' : get_image('tiles/mirrors/bottomleft'),
			'mirror4' : get_image('tiles/mirrors/bottomright')
		}
		for mirror_key in mirror_states.keys():
			loc = ids[mirror_key]
			x = (loc.x << 4) + offset[0]
			y = (loc.y << 4) + offset[1]
			screen.blit(mirror_images[mirror_states[mirror_key]], (x, y))
		
		if get_var('light_puzzle_on') == None:
			return
		
		self.draw_light_beam(screen, 'source', 'A', offset, ids, mirror_states)
		
	def draw_light_beam(self, screen, start, end, offset, ids, mirror_states):
		start_id = ids[start]
		end_id = ids[end]
		start_x = (start_id.x << 4) + 8
		start_y = (start_id.y << 4) + 8
		end_x = (end_id.x << 4) + 8
		end_y = (end_id.y << 4) + 8
		
		pygame.draw.line(screen, (255, 255, 255), (start_x + offset[0], start_y + offset[1]), (end_x + offset[0], end_y + offset[1]))
		
		next = None
		if end == 'A':
			if mirror_states['A'] == 'mirror1':
				next = 'B'
			elif mirror_states['A'] == 'mirror2':
				next = 'G'
		elif end == 'B':
			if mirror_states['B'] == 'mirror2':
				next = 'C'
			elif mirror_states['B'] == 'mirror3':
				next = 'fail1'
		elif end == 'C':
			if mirror_states['C'] == 'mirror3':
				next = 'D'
			elif mirror_states['C'] == 'mirror4':
				next = 'fail2'
		elif end == 'D':
			if mirror_states['D'] == 'mirror1':
				next = 'E'
			elif mirror_states['D'] == 'mirror4':
				next = 'fail8'
		elif end == 'E':
			if mirror_states['E'] == 'mirror3':
				next = 'F'
			elif mirror_states['E'] == 'mirror4':
				next = 'fail9'
		elif end == 'F':
			if mirror_states['F'] == 'mirror1':
				next = 'blue_door'
			elif mirror_states['F'] == 'mirror4':
				next = 'fail10'
		elif end == 'G':
			if mirror_states['G'] == 'mirror1':
				next = 'H'
			elif mirror_states['G'] == 'mirror4':
				next = 'fail3'
		elif end == 'H':
			if mirror_states['H'] == 'mirror3':
				next = 'I'
			elif mirror_states['H'] == 'mirror4':
				next = 'bs_light_to_mainroom_f_puzzle'
		elif end == 'I':
			if mirror_states['I'] == 'mirror1':
				next = 'J'
			elif mirror_states['I'] == 'mirror4':
				next = 'L'
		elif end == 'J':
			if mirror_states['J'] == 'mirror4':
				next = 'K'
			elif mirror_states['J'] == 'mirror3':
				next = 'fail6'
		elif end == 'K':
			if mirror_states['K'] == 'mirror2':
				next = 'red_door'
			elif mirror_states['K'] == 'mirror3':
				next = 'fail7'
		elif end == 'L':
			if mirror_states['L'] == 'mirror1':
				next = 'M'
			elif mirror_states['L'] == 'mirror2':
				next = 'fail4'
		elif end == 'M':
			if mirror_states['M'] == 'mirror2':
				next = 'N'
			elif mirror_states['M'] == 'mirror3':
				next = 'fail5'
		elif end == 'N':
			next = 'yellow_door'
		
		if next != None:
			self.draw_light_beam(screen, end, next, offset, ids, mirror_states)
	
	def light_puzzle_get_mirror_states(self):
		global _defaultMirror
		get_var = ActiveGame().GetVar
		mirrors = {}
		for mirror in 'ABCDEFGHIJKLMN':
			state = get_var('mirror_state_' + mirror)
			if state == None:
				state = _defaultMirror[mirror]
				ActiveGame().SetSavedVar('mirror_state_' + mirror, _defaultMirror[mirror])
			mirrors[mirror] = state
		return mirrors
	
	
	
	def get_camera_offset(self):
		
		width = self.level.width * 16
		height = self.level.height * 16
		
		screen_width = 24 * 16
		screen_height = 18 * 16
		
		offset_x = 0
		offset_y = 0
		
		player_x = self.player.x
		player_y = self.player.y
		
		if width < screen_width:
			offset_x = (screen_width - width) / 2
		elif width > screen_width:
			offset_x = screen_width / 2 - player_x
			offset_x = min(offset_x, 0)
			offset_x = max(offset_x, -(width - screen_width))
			
		if height < screen_height:
			offset_y = (screen_height - height) / 2
		elif height > screen_height:
			offset_y = screen_height / 2 - player_y
			offset_y = min(offset_y, 0)
			offset_y = max(offset_y, -(height - screen_height))
		return (offset_x, offset_y)
	
	def gc_sprites(self):
		sprites = []
		for sprite in self.sprites:
			if not sprite.expired:
				sprites.append(sprite)
		self.sprites = sprites
	
	def get_sprites(self):
		return [self.player] + self.sprites
	
	def get_renderable_sprites(self, layer):
		sprites = []
		for sprite in self.sprites:
			if sprite.layer == layer:
				sprites.append(sprite)
		
		
		
		if self.player.layer == layer and not self.player_invisible:
			
			return [self.player] + sprites
		else:
			return sprites
	
	def do_sprite_move(self, sprite, vx, vy, is_flying_sprite):
		vx = int(vx)
		vy = int(vy)
		#hacked in a lower location for the center of the sprite in the particular case of the player
		params = self.level.move_request(sprite.layer, sprite.x, sprite.y, vx, vy, sprite.r - 4, is_flying_sprite)
		sprite.layer = params[0]
		sprite.x = params[1]
		sprite.y = params[2]
		return params[3]



### .\source\GamePlay\IdMarker.py


class IdMarker:
	def __init__(self, name, layer, x, y, script):
		self.name = name
		self.script = script
		self.layer = layer
		self.y = y
		self.x = x
	

### .\source\GamePlay\KeyRegistry.py

class KeyRegistry:
	def __init__(self):
		self.doors = {}
		
	def RegisterDoor(self, map, dungeon, x, y, color):
		index = self.GetIndexForDoor(map, dungeon, x, y, color)
		above = self.GetIndexForDoor(map, dungeon, x, y - 1, color)
		if self.doors.get(index) == None:
			if self.doors.get(above) == None:
				self.doors[index] = 'registered'
				return True
		return False
		
	def GetIndexForDoor(self, map, dungeon, x, y, color):
		return 'keyregistry_' + '[' + map + ']' + dungeon + '_' + str(x) + '_' + str(y) + '_' + color
	
	def GetPreferredIndexForDoor(self, map, dungeon, x, y, color):
		index = self.GetIndexForDoor(map, dungeon, x, y, color)
		if self.doors.get(index) == None:
			index = self.GetIndexForDoor(map, dungeon, x, y - 1, color)
		return index
	
	def IsDoorLocked(self, map, dungeon, x, y, color):
		return self.GetVar(self.GetPreferredIndexForDoor(map, dungeon, x, y, color)) != 1
	
	def UseKey(self, dungeon, color, map, x, y):
		if self.GetKeyCount(dungeon, color) > 0:
			if self.IsDoorLocked(map, dungeon, x, y, color):
				self.SetVar(self.GetKeyCountIndex(dungeon, color), self.GetKeyCount(dungeon, color) - 1)
				self.SetVar(self.GetPreferredIndexForDoor(map, dungeon, x, y, color), 1)
				play_sound("unlock")
				return True
		return False
	
	def GetKeyCountIndex(self, dungeon, color):
		return 'key_registry_count_' + dungeon + '_' + color
	
	def GetKeyCount(self, dungeon, color):
		index = self.GetKeyCountIndex(dungeon, color)
		count = self.GetVar(index)
		if count == None:
			return 0
		return count
	
	def AddKey(self, dungeon, color):
		index = self.GetKeyCountIndex(dungeon, color)
		count = self.GetVar(index)
		if count == None: count = 0
		self.SetVar(index, count + 1)
	
	def GetVar(self, var):
		return ActiveGame().GetVar(var)
	def SetVar(self, var, value):
		ActiveGame().SetSavedVar(var, value)



### .\source\GamePlay\KeysRenderer.py

class KeyRenderer:
	
	def __init__(self):
		pass
	
	def Render(self, screen, game_scene):
		colors="black blue green indigo red white yellow".split()
		images= [get_image('tiles/keys/%s/key.png'%color) for color in colors]
		currentdungeon=game_scene.level.dungeon
		counts= [GetKeyRegistry().GetKeyCount(currentdungeon,color) for color in colors]
		xoffset=16
		for image, count in zip(images,counts):
			area=pygame.Rect(xoffset,screen.get_clip().bottom-16,8,8)
			if count:
				screen.blit(image, area)
			xoffset+=16
			screen.blit
		
			

### .\source\GamePlay\Layer.py



class Layer:
	def __init__(self, width, height):
		self.contains_stuff = False
		self.width = width
		self.height = height
	def SetTiles(self, tile_list):
		self.tiles = make_table(self.width, self.height)
		self.contains_stuff = True
		
		top = self.height - 1
		bottom = 0
		left = self.width - 1
		right = 0
		
		y = 0
		while y < self.height:
			x = 0
			while x < self.width:
				tl = tile_list[x + y * self.width]
				self.tiles[x][y] = tl
				if not tl.is_blank:
					top = min(y, top)
					bottom = max(y, bottom)
					left = min(x, left)
					right = max(x, right)
				x += 1
			y += 1
		
		self.top = top
		self.bottom = bottom
		self.left = left
		self.right = right

	def MarkStairTile(self, x, y):
		self.tiles[x][y].is_stair_tile()
	
	def GetPhysics(self, x, y, blank_blocked):
		if x >= 0 and x < self.width and y >= 0 and y < self.height:
			
			if not blank_blocked and self.tiles[x][y].is_blank:
				return 'oooo'
			return self.tiles[x][y].physics
		return 'xxxx'
	
	def RunScript(self, x, y):
		if self.contains_stuff:
			if x >= 0 and x < self.width and y >= 0 and y < self.height:
				tile = self.tiles[x][y]
				if tile.id != None:
					script = tile.id.script
					if script != None and script != '':
						go_script_go(script)
	
	def Render(self, screen, x_offset, y_offset, render_counter):
		width = self.width
		height = self.height
		tiles = self.tiles
		
		left = max(self.left, int(x_offset / -16) - 2)
		top = max(self.top, int(y_offset / -16) - 2)
		right = min(left + 24 + 4, self.right + 1)
		bottom = min(top + 18 + 4, self.bottom + 1)
		
		y = top
		while y < bottom:
			x = left
			while x < right:
				tiles[x][y].Render(screen, x * 16 + x_offset, y * 16 + y_offset, render_counter)
				x += 1
			y += 1


### .\source\GamePlay\Level.py


	
class Level:
	def __init__(self, name):
		self.name = name
		get_tile_store()
		self.parse_file(name)
		self.playerStandingOn = None
	
	def synch_stand_key(self, layer, x, y):
		key = layer + str(x) + '_' + str(y)
		if self.playerStandingOn == key:
			self.playerStandingOn = key
			return False
		self.playerStandingOn = key
		return True
	
	def update_tile_standing_on(self, layer, x, y):
		tile_x = x >> 4
		tile_y = y >> 4
		if self.synch_stand_key(layer, tile_x, tile_y):
			layer = self.layers[layer]
			layer.RunScript(tile_x, tile_y)
	
	def parse_file(self, file):
		c = open('maps' + os.sep + file + '.txt', 'rt')
		lines = c.read().split('\n')
		c.close()
		
		values = {}
		for line in lines:
			parts = trim(line).split(':')
			if len(parts) >= 2:
				
				name = parts[0].split('#')[-1]
				value = trim(':'.join(parts[1:]))
				values[name] = value

		self.width = int(values['width'])
		self.height = int(values['height'])
		self.music = values.get('music', None)
		self.layers = {}
		self.dungeon = trim(values.get('dungeon'))
		
		GetKeyRegistry().doors = {}
		
		for layerName in 'A B C D E F Stairs'.split(' '):
			
			content = values.get('Layer' + layerName)
			layer = Layer(self.width, self.height)
			if content != None:
				spots = trim(content).split(',')
				raw_tile_list = []
				for spot in spots:
					tiles = spot.split('|')
					raw_tile_list.append(Tile(tiles))
				layer.SetTiles(raw_tile_list)
			self.layers[layerName] = layer
		
		stair_layer = self.layers['Stairs']
		if stair_layer.contains_stuff:
			other_layers = 'A B C D E F'.split(' ')
			y = 0
			while y < self.height:
				x = 0
				while x < self.width:
					if stair_layer.tiles[x][y].physics != 'xxxx':
						for layerName in other_layers:
							layer = self.layers[layerName]
							if layer.contains_stuff:
								layer.tiles[x][y].is_stair_tile()
					x += 1
				y += 1
		
		kr = GetKeyRegistry()
		self.locked_doors = {}
		self.unique_locked_doors = []
		if self.dungeon != None:
			for layerName in 'A B C D E F'.split(' '):
				layer = self.layers[layerName]
				if layer.contains_stuff:
					y = 0
					while y < self.height:
						x = 0
						while x < self.width:
							tile = layer.tiles[x][y]
							dc = tile.door_color
							if dc != None:
								kr.RegisterDoor(self.name, self.dungeon, x, y, dc)
								for ofs in [(0,1),(0,-1),(1,0),(-1,0)]:
									self.locked_doors[str(x + ofs[0]) + '_' + str(y + ofs[1])] = [x, y, dc]
								self.unique_locked_doors.append([x, y, dc])
								
							x += 1
						y += 1
		
			for locked_door in self.unique_locked_doors:
				x = locked_door[0]
				y = locked_door[1]
				color = locked_door[2]
				if not kr.IsDoorLocked(self.name, self.dungeon, x, y, color):
					self.RemoveLockedDoor(x, y)
	
		script_strings = values.get('scripts')
		scripts = {}
		if script_strings != None:
			scripts_banana = trim(script_strings).split('|||')
			for banana in scripts_banana:
				lines = trim(banana).split('|')
				if len(lines) > 1:
					name = lines[0]
					body = '\n'.join(lines[1:])
					scripts[name] = body
		id_strings = values.get('IDs')
		ids = {}
		if id_strings != None:
			id_strings = trim(id_strings).split(',')
			for id_string in id_strings:
				parts = trim(id_string).split('|')
				if len(parts) == 4:
					name = parts[0]
					layer = parts[1]
					x = int(parts[2])
					y = int(parts[3])
					script = scripts.get(name)
					id = IdMarker(name, layer, x, y, script)
					self.layers[layer].tiles[x][y].SetId(id)
					ids[name] = id
		self.ids = ids
		self.enemies = trim(values.get('enemies', '')).split(',')
		self.on_load = trim(values.get('on_load', '').replace("\\n","\n").replace("\\\\","\\"))
		self.on_enemies_killed = trim(values.get('on_enemies_killed', '').replace("\\n","\n").replace("\\\\","\\"))
		
	def RemoveLockedDoor(self, x, y):
		for layerName in 'A B C D E F'.split(' '):
			layer = self.layers[layerName]
			if layer.contains_stuff:
				if layer.tiles[x][y].door_color != None:
					layer.tiles[x][y].RemoveKey()
					if y > 0:
						layer.tiles[x][y - 1].RemoveKey()
					if y < self.height - 1:
						layer.tiles[x][y + 1].RemoveKey()
				
	def Render(self, layername, screen, x_offset, y_offset, render_counter):
		layer = self.layers[layername]
		if layer.contains_stuff:
			layer.Render(screen, x_offset, y_offset, render_counter)
	
	
	def is_stair_tile(self, x, y):
		stairs = self.layers['Stairs']
		if stairs.contains_stuff and x>=0 and y >= 0 and x < self.width and y < self.height:
			return stairs.tiles[x][y].physics == 'oooo'
		return False
	
	def move_request(self, orig_layer, orig_x, orig_y, dx, dy, radius, is_flying):
		orig_x = int(orig_x)
		orig_y = int(orig_y)
		dest_x = orig_x + dx
		dest_y = orig_y + dy
		
		left = orig_x
		right = orig_x
		top = orig_y
		bottom = orig_y
		
		if orig_x < dest_x:
			right = dest_x
		else:
			left = dest_x
		
		if orig_y < dest_y:
			bottom = dest_y
		else:
			top = dest_y
		
		left -= radius
		top -= radius
		right += radius
		bottom += radius
		
		walls = self.get_walls(orig_layer, left, top, right, bottom, not is_flying)
		
		collided = False
		
		if not self.rectangle_touches_walls(dest_x - radius, dest_y - radius, dest_x + radius, dest_y + radius, walls):
			coords = (dest_x, dest_y)
		else:
			collided = True
			final_x = orig_x
			final_y = orig_y
			
			# check x component
			for x in self.get_between_values(orig_x, dest_x):
				if not self.rectangle_touches_walls(x - radius, final_y - radius, x + radius, final_y + radius, walls):
					final_x = x
					break
			
			# check y component
			for y in self.get_between_values(orig_y, dest_y):
				if not self.rectangle_touches_walls(final_x - radius, y - radius, final_x + radius, y + radius, walls):
					final_y = y
					break
			
			coords = (final_x, final_y)
			
			
		tile_x = coords[0] >> 4
		tile_y = coords[1] >> 4
		if self.is_stair_tile(tile_x, tile_y):
			final_layer = "Stairs"
		else:
			if orig_layer == "Stairs":
				final_layer = 'A' # default value that ought to never get hit.
				# if there's every a bug where you get stuck in the stairs, start looking here
				for layerName in 'A B C D E F'.split(' '):
					layer = self.layers[layerName]
					if layer.contains_stuff:
						if layer.tiles[tile_x][tile_y].physics != 'xxxx':
							final_layer = layerName
							break
					
			else:
				final_layer = orig_layer
			
		return (final_layer, coords[0], coords[1], collided)
	
	def get_between_values(self, start, end):
		values = []
		if end > start:
			for i in range(end, start - 1, -1):
				values.append(i)
		else:
			for i in range(end, start + 1):
				values.append(i)
		return values
	
	def rectangle_touches_walls(self, left, top, right, bottom, walls):
		
		for wall in walls:
			if left > wall[2] + wall[0] or top > wall[1] + wall[3] or right < wall[0] or bottom < wall[1]:
				continue
			return True
		return False
	
	# pixel coordinates
	def get_walls(self, layer, left, top, right, bottom, blank_blocked):
		
		tile_left = (left - 4) >> 4
		tile_right = (right + 4) >> 4
		tile_top = (top - 4) >> 4
		tile_bottom = (bottom + 4) >> 4
		
		walls = []
		
		if layer != 'Stairs':
			layer = self.layers[layer]
			
			y = tile_top
			while y <= tile_bottom:
				x = tile_left
				while x <= tile_right:
					if not self.is_stair_tile(x, y):
						phys = layer.GetPhysics(x, y, blank_blocked)
						if phys == 'oooo':
							pass
						elif phys == 'xxxx':
							walls.append((x << 4, y << 4, 16, 16))
						else:
							if phys[0] == 'x':
								walls.append((x << 4, y << 4, 8, 8))
							if phys[1] == 'x':
								walls.append(((x << 4) + 8, y << 4, 8, 8))
							if phys[2] == 'x':
								walls.append((x << 4, (y << 4) + 8, 8, 8))
							if phys[3] == 'x':
								walls.append(((x << 4) + 8, (y << 4) + 8, 8, 8))
							
					x += 1
				y += 1
		else:
			layerNames = 'A B C D E F'.split(' ')
			y = tile_top
			while y <= tile_bottom:
				x = tile_left
				while x <= tile_right:
					if not self.is_stair_tile(x, y):
						for layerName in layerNames:
							other_layer = self.layers[layerName]
							any_found = False
							if other_layer.contains_stuff:
								phys = other_layer.tiles[x][y].physics 
								if phys == 'oooo':
									any_found = True
								elif phys == 'xxxx':
									pass
								else:
									any_found = True
									if phys[0] == 'x':
										walls.append((x << 4, y << 4, 8, 8))
									if phys[1] == 'x':
										walls.append(((x << 4) + 8, y << 4, 8, 8))
									if phys[2] == 'x':
										walls.append((x << 4, (y << 4) + 8, 8, 8))
									if phys[3] == 'x':
										walls.append(((x << 4) + 8, (y << 4) + 8, 8, 8))
							if any_found:
								break
						if not any_found:
							walls.append((x << 4, y << 4, 16, 16))
						
					x += 1
				y += 1
			
			
		
		return walls
		

### .\source\GamePlay\LevelStore.py



### .\source\GamePlay\OverlayRenderer.py

class OverlayRenderer:
	def __init__(self):
		self.compass = CompassRenderer()
		self.keys = KeyRenderer()
	
	def Render(self, screen):
		self.compass.Render(screen, None)
		self.keys.Render(screen,ActiveGame().GetActiveGameScene())
		
		life = get_life()
		max_life = get_max_life()
		
		
		top = 8
		left = 150
		width = 10
		height = 12
		i = 0
		
		while i < max_life:
			x = left + i * width
			y = top
			pygame.draw.rect(screen, (50, 50, 50), Rect(x, y, width - 1, height))
			pygame.draw.rect(screen, (0, 0, 0),    Rect(x + 1, y + 1, width - 3, height - 2))
			
			if life > i:
				pygame.draw.rect(screen, (255, 0, 0),  Rect(x + 2, y + 2, width - 5, height - 4))
				pygame.draw.rect(screen, (140, 0, 0),  Rect(x + 2, y + 7, width - 5, height - 9))  
			i += 1
		

### .\source\GamePlay\Sprites\Enemy.py

class Enemy:
	def __init__(self, name, id=None):
		self.name = name
		self.id = id
		self.is_enemy = True
		self.explode_on_impact = False
		self.x = 100
		self.y = 100
		self.layer = 'A'
		self.r = 8
		self.direction = 'right'
		self.walking = False
		self.expired = False
		self.state = 'standing'
		self.flying = False
		self.frozen = False
		self.is_goody = False
		self.state_counter = 0
		self.dx = 0
		self.dy = 0
		self.life = 0
		self.flash_counter = -1
		self.game_scene = ActiveGame().GetActiveGameScene()
		if self.name == 'blob':
			self.life = 2
			self.state = 'thinking'
			self.state_counter = int(30 * random.random())
			self.r = 8
		elif self.name == 'eyeball':
			self.life = 2
			self.state = 'walking'
			self.state_counter = int(30 * random.random())
			self.r = 8
		elif self.name == 'mechanicalman':
			self.life = 4
			self.state = 'walking'
			self.state_counter = int(30 * random.random())
			self.r = 8
		elif self.name == 'beetle':
			self.life = 1
			self.state = 'walking'
			self.state_counter = int(30 * random.random())
			self.r = 8
		elif self.name == 'death':
			self.r = 16
			self.life = 15
			self.state = 'walking'
			
			
	
	def DrawingCoords(self):
		offsets = (0,0)
		if self.name == 'mechanicalman':
			offsets = (0, -6)
		elif self.name == 'eyeball':
			offsets = (0, (self.state_counter // 10) & 1)
		elif self.name == 'death':
			offsets = (0, 8)
		coords = (self.x - self.r - offsets[0], self.y - self.r - offsets[1])
		
		return coords
	
	def get_goody(self):
		i = int(random.random() * 10)
		g = None
		if self.name== 'death': return None
		if i < 3:
			g = Goody('life')
		else:
			g = Goody('money')
			
		if g != None:
			g.layer = self.layer
			g.x = self.x
			g.y = self.y
		return g
	
	def Update(self):
		self.dx = 0
		self.dy = 0
		self.state_counter -= 1
		self.flash_counter -= 1
		player_x = self.game_scene.player.x
		player_y = self.game_scene.player.y
		
		delta_x = self.x - player_x
		delta_y = self.y - player_y
		
		dc = None
		for death_circle in self.game_scene.death_circles:
			if death_circle.touches_sprite(self):
				dc = death_circle
				self.flash_counter = 10
				self.life -= 1
		
		if self.life <= 0:
			self.expired = True
			goody = self.get_goody()
			if goody != None:
				self.game_scene.sprites.append(goody)
			return
		
		if not self.frozen:
			if self.name == 'death':
				if player_x < self.x:
					self.dx = -1
				elif player_x > self.x:
					self.dx = 1
				if player_y < self.y:
					self.dy = -1
				elif player_y > self.y:
					self.dy = 1
			elif self.name == 'blob':
				if self.state_counter <= 0:
					if self.state == 'thinking':
						self.state = 'approach'
						self.state_counter = 15
					elif self.state == 'approach':
						self.state = 'thinking'
						self.state_counter = 30
				if self.state == 'approach':
					if player_x > self.x:
						self.dx = 1
					elif player_x < self.x:
						self.dx = -1
					if player_y > self.y:
						self.dy = 1
					elif player_y < self.y:
						self.dy = -1
			elif self.name == 'eyeball' or self.name == 'beetle':
				if self.state_counter <= 0:
					self.state_counter = 50
					if self.name == 'beetle':
						self.state_counter = 50
					self.direction = random.choice('right left down up'.split(' '))
					
				if self.direction == 'left':
					self.dx = -1
				elif self.direction == 'right':
					self.dx = 1
				elif self.direction == 'up':
					self.dy = -1
				else:
					self.dy = 1
			
			elif self.name == 'mechanicalman':
				if self.state_counter <= 0:
					if self.state == 'walking':
						self.state = 'standing'
					elif self.state == 'standing':
						self.state = 'walking'
					self.state_counter = 15
				if self.state == 'walking':
					if abs(delta_x) > abs(delta_y):
						if delta_x > 0:
							self.direction = 'left'
							self.dx = -1
						else:
							self.direction = 'right'
							self.dx = 1
					else:
						if delta_y > 0:
							self.direction = 'up'
							self.dy = -1
						else:
							self.direction = 'down'
							self.dy = 1
			
			
	
	def CurrentImage(self, render_counter):
		if self.frozen: render_counter = 0
		if self.flash_counter > 0 and self.flash_counter & 2 == 0:
			return None
			
		if self.name == 'blob':
			counter = str((render_counter // 4) & 1)
			return get_image('sprites/blob/anim' + counter)
		elif self.name == 'death':
			counter = ('0','1','0','2')[render_counter & 3]
			return get_image('sprites/death/' + counter)
		elif self.name == 'eyeball':
			return get_image('sprites/eyeball/' + self.direction)
		elif self.name == 'beetle':
			counter = str(1 + ((render_counter // 4) & 1))
			f = 'horizontal'
			if self.direction == 'up' or self.direction == 'down':
				f = 'vertical'
			return get_image('sprites/beetle/' + f + counter)
		elif self.name == 'mechanicalman':
			if self.state == 'standing':
				counter = '0'
			else:
				counter = ('0','1','0','2')[render_counter & 3]
			return get_image('sprites/mechanicalman/' + self.direction + str(counter))
		return get_image('sprites/blob/anim0')
		

### .\source\GamePlay\Sprites\Goody.py

class Goody:
	def __init__(self, name, id=None):
		self.name = name
		self.id = id
		self.x = 100
		self.y = 100
		self.dx = 0
		self.dy = 0
		self.is_goody = True
		self.is_enemy = False
		self.layer = 'A'
		self.r = 8
		self.direction = 'right'
		self.flying = False
		self.expired = False
		self.walking = False
		self.state = 'standing'
		self.state_counter = 0
		self.explode_on_impact = False
	
	def DrawingCoords(self):
		return (self.x,self.y)
	
	def Update(self):
		pass
		
	def CurrentImage(self, render_counter):
		counter = render_counter % 30
		if counter < 4:
			counter = '1'
		elif counter < 8:
			counter = '2'
		else:
			counter = '0'
		if self.name == 'money':
			return get_image('misc/money' + counter)
		else:
			return get_image('misc/life_refill' + counter)
		

### .\source\GamePlay\Sprites\NPC.py

class NPC:
	def __init__(self, name, id=None):
		self.name = name
		self.id = id
		self.x = 100
		self.y = 100
		self.dx = 0
		self.dy = 0
		self.is_goody = False
		self.is_enemy = False
		self.layer = 'A'
		self.r = 8
		self.direction = 'right'
		self.flying = False
		self.expired = False
		self.walking = False
		self.state = 'standing'
		self.state_counter = 0
		self.explode_on_impact = False
	
	def DrawingCoords(self):
	
		coords = (self.x - self.r, self.y - self.r - 13)
		if self.name == 'shopkeeper': return (coords[0], coords[1] - 3)
		return coords
	
	def Update(self):
		self.state_counter -= 1
		if self.state_counter <= 0 and self.state != 'walking' and self.state != 'standing':
			self.state = 'standing'
		
	
	def CurrentImage(self, render_counter):
		if self.name == 'shopkeeper':
			return get_image('sprites/townsdude/down0')
		if self.state == 'walking' or self.state == 'standing':
			if self.state == 'walking':
				counter = ('0','1','0','2')[(render_counter // 3) & 3]
			else:
				counter = 0
			counter = str(counter)
			if self.direction == 'right':
				return get_image('sprites/'+self.name+'/right' + counter)
			if self.direction == 'left':
				return get_image('sprites/'+self.name+'/left' + counter)
			if self.direction == 'up':
				return get_image('sprites/'+self.name+'/up' + counter)
			if self.direction == 'down':
				return get_image('sprites/'+self.name+'/down' + counter)
	
		return get_image('sprites/'+self.name+'/down' + counter)
		

### .\source\GamePlay\Sprites\Player.py

_invincible = False

class Player:
	def __init__(self):
		self.x = 100
		self.y = 100
		self.layer = 'A'
		self.r = 8
		self.id = 'MC'
		self.dx = 0
		self.is_goody = False
		self.dy = 0
		self.is_enemy = False
		self.direction = 'right'
		self.walking = False
		self.state = 'walking'
		self.expired = False
		self.flying = False
		self.flash_counter = -1
		self.state_counter = 0
		self.explode_on_impact = False
	
	def DrawingCoords(self):
	
		coords = (self.x - self.r, self.y - self.r - 13)
		if self.state == 'stabbing' or self.state == 'hammering' or self.state == 'drilling':
			if self.direction == 'left':
				coords = (coords[0] - 16, coords[1])
			elif self.direction == 'up':
				coords = (coords[0], coords[1] - 16)
		return coords
	
	def Update(self):
		global _invincible
		self.state_counter -= 1
		self.flash_counter -= 1
		if self.state_counter <= 0:
			self.state = 'walking'
		if self.flash_counter < 0:
			if not _invincible:
				game_scene = ActiveGame().GetActiveGameScene()
				for sprite in game_scene.sprites:
					if sprite.is_enemy:
						if sprite.layer == self.layer:
							dx = sprite.x - self.x
							dy = sprite.y - self.y
							if dx ** 2 + dy ** 2 < (self.r + sprite.r) ** 2:
								self.flash_counter = 30
								take_damage(1)
								if get_life() == 0:
									set_life(3)
									ActiveGame().GetActiveGameScene().next = GameOverScene()
	
	def Stab(self):
		if self.is_submerged(): return
		self.state_counter = 7
		self.state = 'stabbing'
		x = self.x
		y = self.y
		if self.direction == 'left':
			x -= 16
		elif self.direction == 'up':
			y -= 16
		elif self.direction == 'down':
			y += 16
		else:
			x += 16
		game_scene = ActiveGame().GetActiveGameScene()
		
		game_scene.place_death_circle('sword', x, y, 8, 5)
		play_sound('sword')
	def Shoot(self, bullet_type, game_scene):
		if self.is_submerged(): return
		self.state_count = 7
		self.state = 'shooting'
		
		sprites = []
		if bullet_type == 'multi':
			also = []
			if self.direction == 'left' or self.direction == 'right':
				also = ['up','down']
			if self.direction == 'down' or self.direction == 'up':
				also = ['left','right']
			for b in also:
				sprites.append(Projectile(bullet_type, True, self.layer, self.x, self.y, b, game_scene))
		sprites.append(Projectile(bullet_type, True, self.layer, self.x, self.y, self.direction, game_scene))
		for s in sprites:
			game_scene.sprites.append(s)
		play_sound('gunshot')
		
	def Dig(self):
		if self.is_submerged(): return
		self.state_counter = 20
		self.state = 'shovelling'
		play_sound('dig')
	
	def Drill(self):
		if self.is_submerged(): return
		self.state_counter = 15
		self.state = 'drilling'
		x = self.x
		y = self.y
		if self.direction == 'left':
			x -= 16
		elif self.direction == 'up':
			y -= 16
		elif self.direction == 'down':
			y += 16
		else:
			x += 16
		game_scene = ActiveGame().GetActiveGameScene()
		
		game_scene.place_death_circle('drill', x, y, 8, 5)
		play_sound('drill')
	
	def Hammer(self):
		if self.is_submerged(): return
		self.state_counter = 15
		self.state = 'hammering'
		x = self.x
		y = self.y
		if self.direction == 'left':
			x -= 16
		elif self.direction == 'up':
			y -= 16
		elif self.direction == 'down':
			y += 16
		else:
			x += 16
		game_scene = ActiveGame().GetActiveGameScene()
		
		game_scene.place_death_circle('hammer', x, y, 8, 5)
		play_sound('hammer')
	
	def Grapple(self):
		if self.is_submerged(): return
		game_scene = ActiveGame().GetActiveGameScene()
		if game_scene.grapple == None:
			start_x = self.x
			start_y = self.y
			end_x = start_x
			end_y = start_y
			if self.direction == 'up':
				end_y -= 16 * 6
			elif self.direction == 'down':
				end_y += 16 * 6
			elif self.direction == 'left':
				end_x -= 16 * 6
			elif self.direction == 'right':
				end_x += 16 * 6
			
			grapple = Projectile('grapple', True, self.layer, self.x, self.y, self.direction, game_scene)
			grapple.end_x = end_x
			grapple.end_y = end_y
			game_scene.sprites.append(grapple)
			game_scene.grapple = grapple
	
	def is_submerged(self):
		gs = ActiveGame().GetActiveGameScene()
		if gs != None:
			layer = gs.level.layers[self.layer]
			x = self.x >> 4
			y = self.y >> 4
			if x >= 0 and x < layer.width and y >= 0 and y < layer.height:
				value = layer.tiles[x][y].submerged
				return value
	
	def CurrentImage(self, render_counter):
		if self.flash_counter > 0 and (self.flash_counter & 1) == 0:
			return None
		
		counter = '0'
		if self.is_submerged():
			if self.direction == 'right':
				return get_image('sprites/maincharacter/sinkright2')
			if self.direction == 'left':
				return get_image('sprites/maincharacter/sinkleft2')
			if self.direction == 'up':
				return get_image('sprites/maincharacter/sinkup2')
			if self.direction == 'down':
				return get_image('sprites/maincharacter/sinkdown2')
		elif self.state == 'walking':
			if self.walking:
				counter = ('0','1','0','2')[(render_counter // 3) & 3]
			else:
				counter = 0
			counter = str(counter)
			if self.direction == 'right':
				return get_image('sprites/maincharacter/right' + counter)
			if self.direction == 'left':
				return get_image('sprites/maincharacter/left' + counter)
			if self.direction == 'up':
				return get_image('sprites/maincharacter/up' + counter)
			if self.direction == 'down':
				return get_image('sprites/maincharacter/down' + counter)
		elif self.state == 'stabbing' or self.state == 'hammering' or self.state == 'drilling' or self.state == 'shovelling':
			if self.state == 'stabbing':
				img_name = 'stab'
				counter = ('1','2','2','1','1')[(self.state_counter // 2) % 5]
			if self.state == 'drilling':
				img_name = 'drill'
				counter = ((self.state_counter // 2) % 3) + 1
			if self.state == 'hammering':
				img_name = 'hammer'
				counter = ('1','1','2','3','3','3','3','3','3','3','3','3','2','1','1')[self.state_counter % 15]
			if self.state == 'shovelling':
				img_name = 'shovel'
				counter = 2 - ((self.state_counter // 11) % 2)
				self.direction = 'right'
			counter = str(counter)
			if self.direction == 'right':
				return get_image('sprites/maincharacter/'+img_name+'right' + counter)
			if self.direction == 'left':
				return get_image('sprites/maincharacter/'+img_name+'left' + counter)
			if self.direction == 'up':
				return get_image('sprites/maincharacter/'+img_name+'up' + counter)
			if self.direction == 'down':
				return get_image('sprites/maincharacter/'+img_name+'down' + counter)
			
		return get_image('sprites/maincharacter/down' + counter)
		

### .\source\GamePlay\Sprites\Projectile.py

_bulletSwitches = {
	'Fire_Room2' : ['switch'],
	'Fire_Key1' : ['switch_B'],
	'light_rightroom_b1' : ['switch'],
	'light_south_southroom' : ['switch_left','switch_right'],
	'light_southroom_b1' : ['switch'],
	'light_bosshall_f1' : ['left','right'],
	'world_bridge' : ['switch']
}

_grapple_singleton = None
def clear_grapple():
	global _grapple_singleton
	_grapple_singleton = None

def set_grapple(grapple):
	global _grapple_singleton
	_grapple_singleton = grapple

def grapple_exists():
	global _grapple_singleton
	return _grapple_singleton != None

class Projectile:
	def __init__(self, type, friendly, layer, x, y, direction, game_scene):
		self.name = ''
		self.game_scene = game_scene
		self.id = ''
		self.is_enemy = not friendly
		self.x = x
		self.y = y
		self.dx = 0
		self.dy = 0
		self.is_goody = False
		self.grapple_start_x = self.x
		self.grapple_start_y = self.y
		self.grapple_duration = 15
		self.grapple_counter = 0
		self.grappled = False
		
		self.layer = layer
		self.r = 4
		if type == 'grapple':
			self.r = 8
		self.direction = direction
		self.walking = False
		self.expired = False
		self.velocity = 6
		self.state = 'standing'
		self.state_counter = 0
		self.explode_on_impact = True
		self.flying = True
		self.end_x = None
		self.end_y = None
		self.kind = type
	
	def DrawingCoords(self):
	
		coords = (self.x - self.r, self.y - self.r)
		return coords
	
	def Update(self):
		global _bulletSwitches
		tiles = _bulletSwitches.get(self.game_scene.name, [])
		for tile in tiles:
			tile = self.game_scene.level.ids[tile]
			x = abs((tile.x << 4) + 8 - self.x)
			y = abs((tile.y << 4) + 8 - self.y)
			layer = tile.layer
			if self.layer == layer and x < 8 and y < 8:
				self.expired = True
				go_script_go(tile.script)
				break
		
		for sprite in self.game_scene.sprites:
			if sprite.is_enemy:
				dx = sprite.x - self.x
				dy = sprite.y - self.y
				if dx ** 2 + dy ** 2 < (self.r + sprite.r) ** 2:
					
					self.expired = True
					sprite.life -= 1
					if self.kind == 'fire':
						sprite.life -= 1
					elif self.kind == 'ice':
						sprite.life += 1
						sprite.expired = False
						sprite.frozen = True
		
		self.state_counter -= 1
		if self.state_counter <= 0 and self.state != 'walking' and self.state != 'standing':
			self.state = 'standing'
		
		if self.direction == 'left':
			self.dx = -self.velocity
			if self.end_x != None and self.x < self.end_x:
				self.expired = True
		elif self.direction == 'right':
			self.dx = self.velocity
			if self.end_x != None and self.x > self.end_x:
				self.expired = True
		elif self.direction == 'up':
			self.dy = -self.velocity
			if self.end_y != None and self.y < self.end_y:
				self.expired = True
		else:
			self.dy = self.velocity
			if self.end_y != None and self.y > self.end_y:
				self.expired = True
		
		
		
	
	def CurrentImage(self, render_counter):
		if self.kind == 'grapple':
			return get_image('sprites/magnet/' + self.direction)
		elif self.kind == 'ice':
			return get_image('sprites/bullets/ice')
		elif self.kind == 'fire':
			return get_image('sprites/bullets/fire')
		return get_image('sprites/bullets/basic')
		

### .\source\GamePlay\Sprites\SpriteFactory.py



### .\source\GamePlay\Tile.py


class Tile:
	def __init__(self, id_list):
		self.orig_stack = id_list[:]
		self.id = None
		self.is_blank = False
		self.initialize()
		
	def SetTile(self, detaillayer, id):
		index = {'base' : 0, 'baseadorn' : 1, 'basedetail' : 2, 'doodad' : 3, 'doodadadorn' : 4, 'excessive' : 5 }[detaillayer]
		self.orig_stack[index] = id
		self.initialize()
	
	def ChopTheBushes(self):
		i = 0
		found = False
		while i < len(self.orig_stack):
			if self.orig_stack[i] == 'nature20':
				self.orig_stack[i] = ''
				found = True
			i += 1
		if found:
			self.initialize()
	def SmashBoulders(self):
		i = 0
		found = False
		while i < len(self.orig_stack):
			if self.orig_stack[i] == 'boulder1':
				self.orig_stack[i] = ''
				found = True
			i += 1
		if found:
			self.initialize()
	def DrillThrough(self):
		i = 0
		found = False
		while i < len(self.orig_stack):
			if self.orig_stack[i] in '49 50 51 52 124 125 126 127 d49 d50 d51 d52 r49 r50 r51 r52 green49 green50 green51 green52 indigo49 indigo50 indigo51 indigo52'.split(' '):
				self.orig_stack[i] = ''
				found = True
			i += 1
		if found:
			self.initialize()
		
	def RemoveKey(self):
		i = 0
		found = False
		tilestore = get_tile_store()
		while i < len(self.orig_stack):
			id = self.orig_stack[i]
			tt = tilestore.GetTile(id)
			if tt != None and tt.physics.endswith('key'):
				self.orig_stack[i] = ''
				found = True
			i += 1
		if found:
			self.initialize()
	
	def initialize(self):
		tile_stack = []
		no_animations = True
		self.door_color = None
		self.is_grappleable = False
		for id in self.orig_stack:
			if trim(id) != '':
				if id == '63':
					self.is_grappleable = True
				tile = _tileStore.GetTile(id)
				if tile.physics == 'floor':
					tile_stack = []
					no_animations = True
				if tile.physics.endswith('key'):
					self.door_color = tile.physics[:-3]
					self.physics = 'xxxx'
				tile_stack.append(tile)
				if tile.num_images > 1:
					no_animations = False
		self.stack = tile_stack
		if len(tile_stack) == 0:
			self.Render = self._dont_render
			self.is_blank = True
		elif no_animations:
			self.Render = self._static_render
		else:
			self.Render = self._animation_render
		
		self.submerged = False
		self.ice = False
		self.keytype = None
		
		self.composite_physics()
	
	def SetId(self, id):
		self.id = id
	
	def composite_physics(self):
		if len(self.stack) == 0:
			self.physics = 'xxxx'
			return
		
		
		physics = [True, True, True, True]
		for tile in self.stack:
			tphys = tile.physics
			if tphys == 'floor': tphys = 'oooo'
			if len(tphys) == 4 and tphys[0] in 'ox' and tphys[1] in 'ox' and tphys[2] in 'ox' and tphys[3] in 'ox':
				for i in (0,1,2,3):
					physics[i] = physics[i] and tphys[i] == 'o'
				self.submerged = False
			else:
				if tphys == 'water': self.submerged = True
				elif tphys == 'ice': self.ice = True
				
				elif tphys == 'redkey': self.keytype = 'red'
				elif tphys == 'bluekey': self.keytype = 'blue'
				elif tphys == 'greenkey': self.keytype = 'green'
				elif tphys == 'yellowkey': self.keytype = 'yellow'
				elif tphys == 'whitekey': self.keytype = 'white'
				elif tphys == 'blackkey': self.keytype = 'black'
				if self.keytype != None:
					physics = [False, False, False, False]
					break

		self.physics = 'xo'[physics[0]] + 'xo'[physics[1]] + 'xo'[physics[2]] + 'xo'[physics[3]]
	
	def is_stair_tile(self):
		self.physics = 'xxxx'
		self.Render = self._dont_render

	def _animation_render(self, screen, x, y, render_counter):
		for tile in self.stack:
			tile.Render(screen, x, y, render_counter)
	
	def _static_render(self, screen, x, y, render_counter):
		for tile in self.stack:
			screen.blit(tile.images[0], (x, y))
	
	def _dont_render(self, screen, x, y, render_counter):
		pass
			

### .\source\GamePlay\TileStore.py



class TileStore:
	def __init__(self):
		tile_file = 'data' + os.sep + 'tiles.txt'
		c = open(tile_file, 'rt')
		lines = c.read().split('\n')
		c.close()
		
		self.templates = {}
		for line in lines:
			tline = trim(line)
			if len(tline) > 0 and tline[0] != '#':
				parts = tline.split('\t')
				if len(parts) == 4 or len(parts) == 5:
					template = TileTemplate(parts)
					self.templates[template.id] = template
	
	def GetTile(self, id):
		return self.templates.get(id)
			



### .\source\GamePlay\TileTemplate.py


class TileTemplate:
	def __init__(self, parts):
		self.id = parts[0]
		self.physics = parts[2]
		self.imagefiles = parts[3].split('|')
		
		images = []
		for img in self.imagefiles:
			images.append(get_image('tiles' + os.sep + trim(img)))
		self.images = images
		
		if len(parts) == 5:
			self.anim_delay = int(parts[4])
		else:
			self.anim_delay = 4
		self.num_images = len(self.images)
		
	def Render(self, screen, x, y, render_counter):
		if self.num_images == 1:
			screen.blit(self.images[0], (x, y))
		else:
			screen.blit(
				self.images[(render_counter // self.anim_delay) % self.num_images],
				(x, y))


### .\source\GamePlay\TransitionScene.py

class TransitionScene:

	def __init__(self, from_scene, to_level, to_tile, transition_type):
		self.from_scene = from_scene
		self.to_scene = GamePlayScene(to_level, 1, 1)
		self.to_scene.overlayRenderer = None
		self.from_scene.overlayRenderer = None
		self.overlayRenderer = OverlayRenderer()
		dest_tile = self.to_scene.level.ids.get(to_tile)
		self.duration = 30
		if transition_type == WARP_PIXELATE:
			self.duration = 80
		self.max_duration = self.duration + 0.0
		self.temp_screen = None
		self.transition_type = transition_type
		play_music(self.to_scene.level.music)
		
		if dest_tile == None:
			self.next = from_scene
		else:
			self.next = self
			self.to_scene.place_player(dest_tile.layer, dest_tile.x, dest_tile.y)
			self.to_scene.player_invisible = True
			self.from_scene.player_invisible = True
			
	def ProcessInput(self, events):
		pass
		
	def Update(self, game_counter):
		self.duration -= 1
		if self.duration <= 0:
			self.next = self.to_scene
			self.next.overlayRenderer = self.overlayRenderer
			self.to_scene.player_invisible = False
		
	def get_to_screen(self):
		temp = GetTempScreen(self.screen)
		temp.fill((0,0,0))
		self.to_scene.Render(temp)
		return temp
		
	def get_from_screen(self):
		temp = GetTempScreen(self.screen)
		temp.fill((0,0,0))
		self.from_scene.Render(temp)
		return temp
		
	def Render(self, screen):
		self.screen = screen
		temp = GetTempScreen(screen)
		progress = self.duration / self.max_duration
		antiprogress = 1 - progress
		transition = self.transition_type
		h = temp.get_height()
		w = temp.get_width()
		x = int(progress * w)
		y = int(progress * h)
		if transition == WARP_SSCROLL:
			x = 0
			screen.blit(self.get_to_screen(), (x, y))
			screen.blit(self.get_from_screen(), (x, y - h))
		elif transition == WARP_NSCROLL:
			x = 0
			y = -y
			screen.blit(self.get_to_screen(), (x, y))
			screen.blit(self.get_from_screen(), (x, y + h))
		elif transition == WARP_ESCROLL:
			x = x
			y = 0
			screen.blit(self.get_to_screen(), (x, y))
			screen.blit(self.get_from_screen(), (x - w, y))
		elif transition == WARP_WSCROLL:
			x = -x
			y = 0
			screen.blit(self.get_to_screen(), (x, y))
			screen.blit(self.get_from_screen(), (x + w, y))
		elif transition == WARP_PIXELATE or transition == WARP_FADE:
			if progress < .5:
				image = self.get_to_screen()
				amount = progress * 2
			else:
				amount = antiprogress * 2
				image = self.get_from_screen()
			
			if transition == WARP_PIXELATE:
				image = self.pixelate_this(image, 1 - ((1 - amount) ** 2))
			
			screen.fill((0,0,0))
			
			alpha = int(max(0,min(255, 255 * (1 - amount))))
			
			image.set_alpha(alpha)
			screen.blit(image, (0,0))
		self.overlayRenderer.Render(screen)
		
			
	def pixelate_this(self, image, amount):
		minimum_w = 10
		maximum_w = image.get_width()
		minimum_h = 7
		maximum_h = image.get_height()
		
		width = int(minimum_w * amount + maximum_w * (1.0 - amount))
		height = int(minimum_h * amount + maximum_h * (1.0 - amount))
		
		tempScreen = GetTempScreenSize(width, height)
		pygame.transform.scale(image, (width, height), tempScreen)
		
		pygame.transform.scale(tempScreen, (image.get_width(), image.get_height()), image)
		return image
		
		
			
		
	
	



### .\source\ImageLib.py

class ImageLibraryClass:
   def __init__(self):
      self._lib = {} # stores name -> path
      self._pathLib = {} # stores path -> surface
   
   def Add(self, name, path):
      if self._lib.get(name):
         return self.FromFile(self._lib[name])
         
      surf = self.FromFile(path)
      if surf:
         self._lib[name] = path
         return surf
      return None
      

   def Get(self, name):
      p = self._lib.get(name)
      if p:
         return self.FromFile(p)
      return None
   
   def FromFile(self, path):
      surf = self._pathLib.get(path)
      if surf:
         return surf

      if not os.path.exists(path):
         print('Could not load %s' % path)
         return None
      
      surf = pygame.image.load(path)
      if surf:
         self._pathLib[path] = surf
         return surf
      else:
         print('Could not load %s' % path)
         return None



### .\source\ImageLibrary.py


class ImageLibrary:

	def __init__(self):
		self.images = {}
	
	def get_image(self, path):
		img = self.images.get(path)
		if img == None:
			file_path = 'images' + os.sep + path.replace('/', os.sep).replace('\\', os.sep)
			if not file_path.endswith('.png'):
				file_path += '.png'
			
			self.images[path] = pygame.image.load(file_path)
		return self.images[path]



### .\source\Input\InputManager.py


_new_enemies = []

def reserializeEnemies():
	global _new_enemies
	active_game = ActiveGame()
	if active_game != None:
		gs = active_game.GetActiveGameScene()
		if gs != None:
			file = gs.name
			c = open('maps' + os.sep + file  + '.txt', 'rt')
			lines = c.read().split('\n')
			c.close()
			
			output = []
			current = []
			enemies_found = False
			for rline in lines:
				line = trim(rline)
				pieces = line.split(':')
				if pieces[0] != '#enemies':
					output.append(line)
				else:
					value = trim(pieces[1])
					if len(value) > 0:
						enemies_found = True
						current = value.split(',') + _new_enemies
						
			if not enemies_found:
				current = []
			
			current = current + _new_enemies
			_new_enemies = []
			if len(current) > 0:
				
				output.append('#enemies:' + ','.join(current))
			output = '\n'.join(output)
			c = open('maps' + os.sep + file + '.txt', 'wt')
			c.write(output)
			c.close()
			
					
def start_enemy_insertion_session():
	global _new_enemies, _invincible
	_new_enmies = []
	_invincible = True

def insert_enemy(key):
	global _new_enemies
	active_game = ActiveGame()
	if active_game != None:
		gs = active_game.GetActiveGameScene()
		if gs != None:
			x = gs.player.x >> 4
			y = gs.player.y >> 4
			layer = gs.player.layer
			if key == '1':
				type = 'blob'
			elif key == '2':
				type = 'mechanicalman'
			elif key == '3':
				type = 'eyeball'
			elif key == '4':
				type = 'beetle'
			else:
				type = None
			
			if type != None:
				_new_enemies.append(type + '|' + layer + '|' + str(x) + '|' + str(y))
				sprite = Enemy(type)
				sprite.x = (x << 4) + 8
				sprite.y = (y << 4) + 8
				sprite.layer = layer
				gs.sprites.append(sprite)
			

class InputEvent:
	def __init__(self, key, down):
		self.key = key
		self.down = down
		self.up = not down
	
	def Up(self):
		return self.key == 'up'
	def Down(self):
		return self.key == 'down'
	def Right(self):
		return self.key == 'right'
	def Left(self):
		return self.key == 'left'
	def A(self):
		return self.key == 'A'
	def B(self):
		return self.key == 'B'
	def X(self):
		return self.key == 'X'
	def Y(self):
		return self.key == 'Y'
	def Start(self):
		return self.key == 'start'
	
	def __str__(self):
		if self.up:
			p = 'not '
		else:
			p = ''
		return 'key: %s, %spressed' % (str(self.key), p)

class Joystick:
	def __init__(self, pygameJoystick):
		self.js = pygameJoystick
		self.name = self.js.get_name()
		
		self.config = {
			'start' : None,
			'A' : None,
			'B' : None,
			'Y' : None,
			'X' : None,
			'up' : None,
			'down' : None,
			'left' : None,
			'right' : None
		}
		
		self.is_pressed = {
			'start' : False,
			'A' : False,
			'B' : False,
			'X' : False,
			'Y' : False,
			'up' : False,
			'down' : False,
			'left' : False,
			'right' : False
		}
		self.prev_config = None
		
	# returns None when state doesn't change
	# returns True when key is pushed
	# returns False when key is released
	def state_changed(self, key):
		config = self.config[key]
		if config == None: return None
		
		pressed = False
		if config[0] == 'hat':
			hat_status = self.js.get_hat(config[1])
			if hat_status[0] == 1 and config[2] == 'right':
				pressed = True
			elif hat_status[0] == -1 and config[2] == 'left':
				pressed = True
			elif hat_status[1] == 1 and config[2] == 'up':
				pressed = True
			elif hat_status[1] == -1 and config[2] == 'down':
				pressed = True
			else:
				pressed = False
		elif config[0] == 'axis':
			axis_status = self.js.get_axis(config[1])
			if config[2]:
				if axis_status > .5:
					pressed = True
			else:
				if axis_status < -.5:
					pressed = True
		elif config[0] == 'button':
			pressed = self.js.get_button(config[1])
			
		if pressed != self.is_pressed[key]:
			self.is_pressed[key] = pressed
			return pressed
		return None
		
	
	def configure_key(self, keyname):
		key = self.get_first_pressed()
		if key != None and key != self.prev_config:
			self.config[keyname] = key
			self.prev_config = key
			return True
		return False
	
	def get_first_pressed(self):
		for hat_num in range(self.js.get_numhats()):
			hat = self.js.get_hat(hat_num)
			if hat != (0,0):
				if hat[0] == -1:
					return ('hat', hat_num, 'left')
				elif hat[0] == 1:
					return ('hat', hat_num, 'right')
				if hat[1] == -1:
					return ('hat', hat_num, 'down')
				elif hat[1] == 1:
					return ('hat', hat_num, 'up')
		for but_num in range(self.js.get_numbuttons()):
			button = self.js.get_button(but_num)
			if button:
				return ('button', but_num)
		
		for ax_num in range(self.js.get_numaxes()):
			axis = self.js.get_axis(ax_num)
			if axis > .5:
				return ('axis', ax_num, 1)
			elif axis < -.5:
				return ('axis', ax_num, -1)
		return None
					
	
_allowEnemyEdit = True
_enemyEditMode = False

class InputManager:
	
	def __init__(self):
		self.key_pressed_now = ''
		
		self.is_pressed = {
			'up' : False,
			'down' : False,
			'left' : False,
			'right' : False,
			'start' : False,
			'A' : False,
			'B' : False,
			'Y' : False,
			'X' : False
		}
		self.escape_attempted = False
		self.active_joystick = None
		
		self.keys = 'up down left right start A B X Y'.split(' ')
		self.video_mode = None
		
	def initializeJoystick(self):
		
		joysticks = []
		for i in range(pygame.joystick.get_count()):
			joysticks.append(Joystick(pygame.joystick.Joystick(i)))
		
		self.joysticks = joysticks
		if len(joysticks) >= 1:
			self.active_joystick = self.joysticks[0]
	
	def VideoModeChange(self):
		return self.video_mode
	
	def get_events(self):
		
		global _allowEnemyEdit, _enemyEditMode
		
		self.key_pressed_now = ''
		events = []
		joystick = self.active_joystick
		if joystick != None:
			for key in self.keys:
				state = joystick.state_changed(key)
				if state != None:
					events.append(InputEvent(key, state))
		
		self.video_mode = None
		
		for event in pygame.event.get():
			if event.type == QUIT or (event.type == KEYDOWN and event.key == K_ESCAPE):
				self.escape_attempted = True
			elif event.type == KEYDOWN:
				if event.key == K_UP:
					events.append(InputEvent('up', True))
				elif event.key == K_DOWN:
					events.append(InputEvent('down', True))
				elif event.key == K_RIGHT:
					events.append(InputEvent('right', True))
				elif event.key == K_LEFT:
					events.append(InputEvent('left', True))
				elif event.key == K_SPACE:
					events.append(InputEvent('B', True))
				elif event.key == K_a:
					events.append(InputEvent('A', True))
				elif event.key == K_s:
					events.append(InputEvent('Y', True))
				elif event.key == K_d:
					events.append(InputEvent('X', True))
				elif event.key == K_RETURN:
					events.append(InputEvent('start', True))
				elif event.key == K_f:
					self.video_mode = 'full'
				elif event.key == K_w:
					self.video_mode = 'wide'
				if _allowEnemyEdit:
					if event.key == K_e:
						_enemyEditMode = not _enemyEditMode
						if _enemyEditMode:
							start_enemy_insertion_session()
						else:
							reserializeEnemies()
							
					elif _enemyEditMode:
						if event.key == K_1:
							insert_enemy('1')
						elif event.key == K_2:
							insert_enemy('2')
						elif event.key == K_3:
							insert_enemy('3')
						elif event.key == K_4:
							insert_enemy('4')
						elif event.key == K_5:
							insert_enemy('5')
						elif event.key == K_6:
							insert_enemy('6')
						elif event.key == K_7:
							insert_enemy('7')
						elif event.key == K_8:
							insert_enemy('8')
						elif event.key == K_9:
							insert_enemy('9')
						elif event.key == K_0:
							insert_enemy('0')
						
			elif event.type == KEYUP:
				if event.key == K_UP:
					events.append(InputEvent('up', False))
				elif event.key == K_DOWN:
					events.append(InputEvent('down', False))
				elif event.key == K_RIGHT:
					events.append(InputEvent('right', False))
				elif event.key == K_LEFT:
					events.append(InputEvent('left', False))
				elif event.key == K_SPACE:
					events.append(InputEvent('B', False))
				elif event.key == K_a:
					events.append(InputEvent('A', False))
				elif event.key == K_s:
					events.append(InputEvent('Y', False))
				elif event.key == K_d:
					events.append(InputEvent('X', False))
				elif event.key == K_RETURN:
					events.append(InputEvent('start', False))
			
		for event in events:
			self.is_pressed[event.key] = event.down
		
		return events
				




### .\source\JukeBox.py

class JukeBox:
	def __init__(self):
		self.now_playing = None
		self.sound_library = {}
		self.current_volume = True
		pygame.mixer.init()
		pygame.mixer.music.set_volume(1.0)
	
	def PlaySong(self, name):
		#return
		if name == 'STOP': name = None
		if not name:
			pygame.mixer.music.stop()
			return
		
		if self.now_playing != name:
			self.MakeMusicLoud()
			self.now_playing = name
			if not name.endswith('.mp3'):
				name=name+".mp3"
			pygame.mixer.music.load('media' + os.sep + 'music' + os.sep + name)
			pygame.mixer.music.play(-1)
	
	def PlaySound(self, name):
		if not name:
			return
		sound = self.sound_library.get(name)
		if sound == None:
			if not name.endswith('.wav'):
				name=name+".wav"
			sound = pygame.mixer.Sound('media' + os.sep + 'sound' + os.sep + name)
			self.sound_library[name] = sound
		sound.play()
	
	def MakeMusicSoft(self):
		if self.current_volume:
			pygame.mixer.music.set_volume(0.5)
			self.current_volume = False
	
	def MakeMusicLoud(self):
		if not self.current_volume:
			pygame.mixer.music.set_volume(1.0)
			self.current_volume = True
		


### .\source\MapScript.py

class MapScript(Scripted):
   # script -- ScriptIter
   def __init__(self, scriptIter):
      Scripted.__init__(self, scriptIter)
      self._addFn('warp', do_warp)
      self._addFn('dialog', do_dialog)
      self._addFn('remove tile', do_removeTile)
      self._addFn('set tile', do_setTile)
      self._addFn('sound', do_sound)
      self._addFn('music', do_music)
      self._addFn('launch cutscene', do_cutscene)
      self._addFn('toggle mirror', do_toggle_mirror)
      self._addFn('get key', do_getkey)
      self._addFn('save', do_save_point)
      self._addFn('sign', do_sign_display)
   
   # get the next bit of stuff to display
   def Advance(self):
      return Scripted.Advance(self)
   
   # fancy name so it makes more sense
   def Exec(self):
      return self.Advance()
   
   def _parseInternal(self, line):
      print('MapScript._parseInternal(%s)' % line)

   # function implementations
   # return indicates if script execution should continue (True) or stop until
   # the next Advance (False)
   def _end(self):
      # what to do?
      return Scripted._end(self)
   
   

### .\source\Scenes\BaseScene.py

class BaseScene:
   def __init__(self):
      self.next = self
   
   def ProcessInput(self, events):
      pass
   
   def Update(self, conter):
      pass
   
   def Render(self, screen):
      pass
   

### .\source\Scenes\Credits.py

class CreditsScene:
   def __init__(self):
      self.next = self
      self._y = 250
      
      lines = [
         "heading1:%s" % GAME_NAME,
         "heading:Team Nerdparadise",
         "www.nerdparadise.com",
         "",
         "",
         "heading:Programming",
         "Blake O'Hare",
         "Richard Bailey",
         "Adrian Cline",
         "",
         "",
         "",
         "heading:Pixel Art",
         "Angel McLaughlin",
         "",
         "",
         "",
         "heading:Character Art",
         "Chun Y",
         "",
         "",
         "",
         "heading:Music",
         "Adrian Cline",
         "",
         "",
         "",
         "heading:Level Design",
         "Christine Sandquist",
         "Yan Yan",
         "Brett S."
         "",
         "",
         "",
         "heading:Special Thanks...",
         "www.anke-art.de, for Fortunaschwein font",
         "Ray Meadows, for RM Typerighter font",
         "",
         "",
         "",
         "heading:Welcoming To Team",
         "heading:Nerdparadise",
         "Christine Sandquist",
         "Yan Yan",
         "Chun Y",
      ]
      
      cSurf = []
      max_x = 0
      cum_y = 0
      for l in lines:
         sz = 18
         if l.startswith('heading1:'):
            l = l.split(':')[1]
            sz = 45
         if l.startswith('heading:'):
            l = l.split(':')[1]
            sz = 30
         surf = render_text_size(sz, l, WHITE)
         cSurf.append(surf)
         max_x = max(max_x, surf.get_width())
         cum_y += surf.get_height()
      
      self._credits = pygame.Surface((max_x, cum_y))
      
      y = 0
      for s in cSurf:
         x = int((max_x - s.get_width()) / 2)
         self._credits.blit(s, (x, y))
         y += s.get_height()
   
   def ProcessInput(self, events):
      for e in events:
         if e.down:
            if e.Start() or e.A() or e.B():
               mm = MainMenuScene()
               self.next = mm
               mm.next = mm
   
   def Update(self, conter):
      pass
   
   def Render(self, screen):
      c = self._credits
      sw = screen.get_width()
      offy = self._y
      
      screen.blit(c, (int((sw - c.get_width()) / 2), offy))
      
      self._y -= 1
      
      if (c.get_height() - abs(self._y)) == -40:
         mm = MainMenuScene()
         self.next = mm
         mm.next = mm


### .\source\Scenes\DialogScene.py

class DialogScene:
   def __init__(self, dlg, sourceScene):
      if not isinstance(dlg, Dialog):
         raise Exception("dlg must an object of type Dialog")
      ImageLib.Add('d-frame', uiImgPath('dframe'))
      
      self.next = self
      self._source = sourceScene
      self._dlg = dlg
      self._choice = 0
      self._tick = 0
      self._fin = False
      
   def ProcessInput(self, events):
      d = self._dlg
      
      if 0 != len(events):
         for e in events:
            if e.down:
               if e.Up() and d.State() == D_QUESTION:
                  self._choice -= 1
                  self._choice %= len(self._dlg.Choices())
               
               if e.Down() and d.State() == D_QUESTION:
                  self._choice += 1
                  self._choice %= len(self._dlg.Choices())
            else:
               if e.A() or e.B():
                  if self._dlg.State() == D_QUESTION:
                     self._dlg.Answer(self._choice)
                  self._dlg.Advance()

   def Update(self, game_counter):
      pass

   def Render(self, screen):
      self._source
      d = self._dlg
      
      if d.State() == D_END:
         self.next = self._source
         self._source.next = self._source

      self._source.Render(screen)
      
      p = d.Profile()
      pSurf = None

      if p:
         side = 'right'
         if p == 'mc_portrait':
            side = 'left'
         p = portraitPath(p)
         if p:
            pSurf = ImageLib.FromFile(p)
      if pSurf:
         if side == 'left':
            screen.blit(pSurf, (4, 110))
         else:
            screen.blit(pSurf, (290, 110))
         
      df = ImageLib.Get('d-frame')
      screen.blit(df, (0,screen.get_height() - df.get_height() - 4))

      txt = d.Text()
      txt = txt.split('\n')
      tSurf = []
      for t in txt:
         if t == '$nl$':
            t = ''
         tSurf.append(_font.render(t, True, BLACK))
      
      lineNo = 0
      for t in tSurf:
         screen.blit(t, (D_TEXT_OFFSET_X, D_TEXT_OFFSET_Y + lineNo * _font.get_height()))
         lineNo += 1

      if D_QUESTION == d.State():
         cy = int(D_TEXT_OFFSET_Y + (lineNo + self._choice) * _font.get_height() + (.5 * _font.get_height()))
         cx = D_TEXT_OFFSET_X + 6
         
         # draw choice indicator
         pygame.draw.circle(screen, BLUE, (cx, cy), 4)
         
         # print choice text
         for c in d.Choices():
            cSurf = _font.render(c, True, BLACK)
            screen.blit(cSurf, (D_ANSWER_OFFSET_X, D_TEXT_OFFSET_Y + lineNo * _font.get_height()))
            lineNo += 1
         

### .\source\Scenes\GameOver.py

class GameOverScene:
	def _LastSave(self):
		ActiveGame().parse()
		m = ActiveGame().GetVar('save_map')
		if m == None:
			m = 'transport_1'
			x = 64
			y = 64
		x = ActiveGame().GetVar('save_x')
		y = ActiveGame().GetVar('save_y')
		if x == None: x = 0
		if y == None: y = 0
		self.next = GamePlayScene(m, x, y + 16)
		self.next.level.update_tile_standing_on(self.next.player.layer, self.next.player.x, self.next.player.y)
		self.next.prevTile = self.next.level.playerStandingOn
		self.next.firstTimeOnTile = False
		self.next.disable_save = True
	
	def _MainMenu(self):
		mm = MainMenuScene()
		self.next = mm
		mm.next = mm
	
	def _Quit(self):
		self.next = None

	def __init__(self):
		self.next = self
		
		self._selection = 0
		self._fc = 0
		self._frame = 0
	
		self._gears = []
		i = 1
		while i <= 4:
			self._gears.append(ImageLib.FromFile(uiImgPath('gear%d' % i)))
			i += 1

	
	def ProcessInput(self, events):
		for e in events:
			if e.down:
				if e.Up():
					self._selection -= 1
					self._selection %= 3
				if e.Down():
					self._selection += 1
					self._selection %= 3
				if e.A() or e.B() or e.Start():
					if self._selection == 0:
						self._LastSave()
					if self._selection == 1:
						self._MainMenu()
					if self._selection == 2:
						self._Quit()
	
	def Update(self, conter):
		pass
	
	def Render(self, screen):
		self._fc += 1
		
		frame = self._frame
		if (self._fc % 2 == 0):
			self._frame += 1
			self._frame %= len(self._gears)

		screen.fill(BLACK)
		sw = screen.get_width()
		
		death = render_text_size(23, "You Have Perished", WHITE)
		death_sub = render_text_size(15,  "The empire has lost its best hope", WHITE)
		death_sub2 = render_text_size(15, "for salvation. Are there any that", WHITE)
		death_sub3 = render_text_size(15, "may hope to fill your place?", WHITE)

		death_x = int((sw - death.get_width()) / 2)
		death_y = 23
		screen.blit(death, (death_x, death_y))
		death_x = int((sw - death_sub.get_width()) / 2)
		death_y = death_y + death.get_height() + 4
		screen.blit(death_sub, (death_x, death_y))
		death_x = int((sw - death_sub2.get_width()) / 2)
		death_y = death_y + death_sub.get_height()
		screen.blit(death_sub2, (death_x, death_y))
		death_x = int((sw - death_sub3.get_width()) / 2)
		death_y = death_y + death_sub2.get_height()
		screen.blit(death_sub3, (death_x, death_y))
		
		cont = render_text_size(23, "Go to Last Save", WHITE)
		main = render_text_size(23, "Main Menu", WHITE)
		quit = render_text_size(23, "Quit", RED)
		
		cx = 135 #int((sw - cont.get_width()) / 2)
		cy = death_y + death_sub3.get_height() + 20
		mx = cx #int((sw - main.get_width()) / 2)
		my = cy + cont.get_height() + 15
		qx = cx #int((sw - quit.get_width()) / 2)
		qy = my + main.get_height() + 13
		
		screen.blit(cont, (cx, cy))
		screen.blit(main, (mx, my))
		screen.blit(quit, (qx, qy))
		
		gx = cx - 50
		gy = cy + (43 * self._selection) - 7
		g = self._gears[frame]
		screen.blit(g, (gx, gy))

### .\source\Scenes\GameSelect.py

class GameSelectScene:
	def __init__(self, prevScene):
		self._prevScene = prevScene
		self.next = self
		
		self._eraseMode = False
		self._erase = False
		self._cancel = False
		self._selection = 0
		self._fc = 0
		self._frame = 0
		self._gears = []
		i = 1
		while i <= 4:
			self._gears.append(ImageLib.FromFile(uiImgPath('gear%d' % i)))
			i += 1
	
	def ProcessInput(self, events):
		for e in events:
			if e.down:
				if self._erase or self._cancel:
					if e.Up():
						play_sound("menu2")
						self._erase = False
						self._cancel = False
						self._selection = 2
						return
					if e.Down():
						play_sound("menu2")
						self._erase = False
						self._cancel = False
						self._selection = 0
						return
					if e.Right() or e.Left():
						self._erase = not self._erase
						self._cancel = not self._cancel
						return
				elif e.Up():
					play_sound("menu2")
					self._selection -= 1
					self._selection %= 4
					
				elif e.Down():
					play_sound("menu2")
					self._selection += 1
					self._selection %= 4
				
				if (e.Up() or e.Down()) and self._selection == 3:
					self._erase = True

				if e.A() or e.B() or e.Start():
					if self._cancel:
						self.next = self._prevScene
						self._prevScene.next = self._prevScene
						return
					
					if self._erase:
						self._eraseMode = True
						return
					
					if self._eraseMode:
						GameContext().DeletePlayer(self._selection + 1)
						self._eraseMode = False
						return
					
					newGame = not GameContext().GetPlayerName(self._selection + 1)
					GameContext().SetActiveGame(self._selection + 1)
					
					if newGame:
						self.next = NameEntryScene()
					else:
						#print('TODO: set up to resume gameplay')
						m = ActiveGame().GetVar('save_map')
						if m == None:
							m = 'transport_1'
							x = 64
							y = 64
						x = ActiveGame().GetVar('save_x')
						y = ActiveGame().GetVar('save_y')
						if x == None: x = 0
						if y == None: y = 0
						self.next = GamePlayScene(m, x, y + 16)
						self.next.level.update_tile_standing_on(self.next.player.layer, self.next.player.x, self.next.player.y)
						self.next.prevTile = self.next.level.playerStandingOn
						self.next.firstTimeOnTile = False
						self.next.disable_save = True
						
	def Update(self, conter):
		play_music("menuwaitingroom")
	
	def Render(self, screen):
		# what color is used for everything else
		txtColor = WHITE
		# what color is in the game select thing
		descColor = BLACK
		# game selection block
		gameSelColor = WHITE
		
		self._fc += 1
		frame = self._frame
		if (self._fc % 2 == 0):
			self._frame += 1
			self._frame %= len(self._gears)
		g = self._gears[frame]

		gc = GameContext()
		
		sw = screen.get_width()
		gameSel_y = 20
		
		gameTxt = "Game Select"
		if self._eraseMode:
			gameTxt = 'Erase Game'
		gameSel = render_text_size(23, gameTxt, txtColor)
		screen.blit(gameSel, (int((sw - gameSel.get_width()) / 2), gameSel_y))
	
		post_title_y = gameSel_y + gameSel.get_height() + 10
	
		# game sel width
		sel_w = int(.7 * sw)
		# game sel height
		sel_h = 45
		# separation between game sel blocks
		gameSel_sep = 15
		selSurf = pygame.Surface((sel_w, sel_h))
		selSurf.set_alpha(ALPHA)
		selSurf.fill(gameSelColor)
		
		i = 0
		while i < 3:
			sx = int(sw * .2)
			sy = i * (gameSel_sep + sel_h) + post_title_y

			if i == self._selection:
				screen.blit(g, (sx - 50, sy))
	
			screen.blit(selSurf, (sx, sy))
			
			slot_num = i + 1
			name = gc.GetPlayerName(slot_num)
			if name:
				nameSurf = render_text_size(17, name, descColor)
			else:
				nameSurf = render_text_size(17, 'New Game', descColor)
			
			nx = sx + 10 #int((selSurf.get_width() - nameSurf.get_width())/2) + sx
			ny = int((selSurf.get_height() - nameSurf.get_height()) / 2) + sy
			screen.blit(nameSurf, (nx, ny))

			if name:
				#stSurf = ImageLib.FromFile(uiImgPath('badges'))
				#bx = sx + selSurf.get_width() - stSurf.get_width()
				#by = int((selSurf.get_height() - stSurf.get_height()) / 2) + sy
				#screen.blit(stSurf, (bx, by))

				stones = gc.GetStones(slot_num)
				j = 0
				stones.reverse()
				for st in stones:
					stSurf = ImageLib.FromFile(uiImgPath('stones', st))
					stx = sx + selSurf.get_width() - 10 - ((1 + j) * 20)
					sty = sy + int((selSurf.get_height() - stSurf.get_height()) / 2)
					
					screen.blit(stSurf, (stx, sty))
					j += 1

			#end while
			i += 1
		
		erase = render_text_size(20, 'Erase', txtColor)
		cancel = render_text_size(20, 'Cancel', txtColor)
		
		screen.blit(erase, (75, 240))
		cx = 230
		screen.blit(cancel, (cx, 240))
		
		gy = 230
		if self._erase:
			screen.blit(g, (sx - 50, gy))
		if self._cancel:
			screen.blit(g, (cx + 10 + cancel.get_width(), gy))
			
		# stone badges
		
		


### .\source\Scenes\Inventory.py

class InventoryScene:
   def __init__(self, overlay):
      self.next = self
      
      i = Inventory()
      self._layout = ([i.Sabre(), i.Hammer(), i.Drill(), i.Hook(), i.Compass()],
                      [i.Cannon(), i.CannonFire(), i.CannonIce(), i.CannonMulti(), i.Shovel()])
      
      self._i = i
      self._baseScene = overlay
      self._itemSurf = {}
      self._selection = [0, 0]
      
      il = ImageLib
      isurf = self._itemSurf
      
      isurf['cannon-icon'] = il.FromFile(uiImgPath('cannon-icon'))
      isurf[i.Cannon()] = il.FromFile(uiImgPath('ammo-sshot'))
      isurf[i.CannonFire()] = il.FromFile(uiImgPath('ammo-fire'))
      isurf[i.CannonIce()] = il.FromFile(uiImgPath('ammo-ice'))
      isurf[i.CannonMulti()] = il.FromFile(uiImgPath('ammo-multi'))
      isurf[i.Sabre()] = il.FromFile(uiImgPath('sabre-have'))
      isurf[i.Hammer()] = il.FromFile(uiImgPath('hammer-have'))
      isurf[i.Drill()] = il.FromFile(uiImgPath('drill-have'))
      isurf[i.Hook()] = il.FromFile(uiImgPath('hook-have'))
      isurf[i.Compass()] = il.FromFile(uiImgPath('compass-have'))
      isurf[i.Shovel()] = il.FromFile(uiImgPath('shovel-have'))
   
   def ProcessInput(self, events):
      for e in events:
         if e.down and e.key == 'start':
            self._baseScene.next = self._baseScene
            self.next = self._baseScene
            # self._i.PrintEquipped()
            return
         if e.down:
            if e.key == 'up':
               self._selection[1] -= 1
               self._selection[1] %= 2
            elif e.key == 'down':
               self._selection[1] += 1
               self._selection[1] %= 2
            elif e.key == 'right':
               self._selection[0] += 1
               self._selection[0] %= 5
            elif e.key =='left':
               self._selection[0] -= 1
               self._selection[0] %= 5
            else:
               self._select(e.key, self._selection[0], self._selection[1])

   def _select(self, key, row, col):
      item = self._layout[col][row]

      i = self._i
      if not i.Check(item):
         print('player does not have %s' % item)
      else:
         i.Equip(key.lower(), item)

   def Update(self, conter):
      pass
   
   def Render(self, screen):
      self._baseScene.Render(screen)
      i = self._i
      isurf = self._itemSurf
      sw = screen.get_width()
      
      
      titleSurf = pygame.Surface((200, 25))
      titleSurf.set_alpha(120)
      
      moneySurf = pygame.Surface((140, 25))
      moneySurf.set_alpha(120)
      
      itemSurf = pygame.Surface((140, 60))
      itemSurf.set_alpha(120)
      
      ss = 1 # strokeSize
      vBorder = 7
      
      money_off_x = int((screen.get_width() - moneySurf.get_width()) / 2)
      money_off_y = 150
      pygame.draw.rect(screen, WHITE, pygame.Rect(money_off_x - ss, money_off_y - ss, 140 + (2 *ss), 25 + (2 *ss)), ss)
      
      item_off_x = int((screen.get_width() - itemSurf.get_width()) / 2)
      item_off_y = money_off_y + titleSurf.get_height() + vBorder
      pygame.draw.rect(screen, WHITE, pygame.Rect(item_off_x - ss, item_off_y - ss, 140 + (2 *ss), 60 + (2 *ss)), ss)
      
      title_off_x = int((screen.get_width() - titleSurf.get_width()) / 2)
      title_off_y = item_off_y + itemSurf.get_height() + vBorder
      pygame.draw.rect(screen, WHITE, pygame.Rect(title_off_x - ss, title_off_y - ss, 200 + (2 *ss), 25 + (2 *ss)), ss)
      
      screen.blit(titleSurf, (title_off_x, title_off_y))
      screen.blit(itemSurf, (item_off_x, item_off_y))
      screen.blit(moneySurf, (money_off_x, money_off_y))

      if i.HasAny():
         # TODO: this is ugly and terrible, fix it
         a = render_text_size(17, 'A', WHITE)
         aw = a.get_width()
         ah = a.get_height()
         b = render_text_size(17, 'B', WHITE)
         bw = b.get_width()
         bh = b.get_height()
         x = render_text_size(17, 'X', WHITE)
         xw = x.get_width()
         xh = x.get_height()
         y = render_text_size(17, 'Y', WHITE)
         yw = y.get_width()
         yh = y.get_height()
         eq = pygame.Surface((aw+bw+40+18, 46))
         eq.set_alpha(120)
         eq.fill(BLACK)
         ex = int((sw - eq.get_width())/2)
         ey = 50
         screen.blit(eq, (ex, ey))
         pygame.draw.rect(screen, WHITE, pygame.Rect(ex - ss, ey - ss, eq.get_width() + (2 * ss), eq.get_height() + (2 * ss)), ss)
         '''
         a    x
         b    y
         '''
         surf = i.Surf('a')
         screen.blit(a, (ex + 3, ey + 0))
         if surf:
            screen.blit(surf, (ex + aw + 6, ey))
         
         surf = i.Surf('b')
         screen.blit(b, (ex + 3, ey + 23))
         if surf:
            screen.blit(surf, (ex + bw + 6, ey + 23))
         
         surf = i.Surf('x')
         screen.blit(x, (ex + aw + 26, ey + 0))
         if surf:
            screen.blit(surf, (ex + xw + aw + 26 + 3, ey))
         
         surf = i.Surf('y')
         screen.blit(y, (ex + bw + 26, ey + 23)) 
         if surf:
            screen.blit(surf, (ex + bw + yw + 26 + 3, ey + 23))

      # draw player money:
      coinSurf = ImageLib.FromFile(os.path.join('images', 'misc', 'money0.png'))
      cx = money_off_x + 4 # moneySurf.get_width() - coinSurf.get_width() - 5
      cy = money_off_y + int((moneySurf.get_height() - coinSurf.get_height()) / 2)
      screen.blit(coinSurf, (cx, cy))
      
      amt = render_number(get_money(), WHITE)
      mx = money_off_x + moneySurf.get_width() - 5 - amt.get_width()
      my = money_off_y + int((moneySurf.get_height() - amt.get_height()) / 2)
      screen.blit(amt, (mx, my))

      # draw all items the player has:
      col = 0
      while col < 5:
         row = 0
         while row < 2:
            item = self._layout[row][col]
            if isurf[item]:
             if i.Check(item):
               screen.blit(isurf[item], (item_off_x + 4 + (28 * col),
                                          item_off_y + 4 + (28 * row)))
            
            row += 1
         col += 1
      
      # deal with selected item label:
      item = self._layout[self._selection[1]][self._selection[0]]
      if i.Check(item):
         text = render_text_size(17, i.Description(item), WHITE)
         
         if text:
            txt_off_x = title_off_x + int((titleSurf.get_width() - text.get_width()) / 2)
            txt_off_y = title_off_y + 2
            
            screen.blit(text, (txt_off_x, txt_off_y))
      
      # draw cannon overlay
      try:
         idx = self._layout[1].index(i.WhichCannonEquipped())
         cannonSurf = self._itemSurf['cannon-icon']
         cannon_off_x = item_off_x + (28 * idx) + 8
         cannon_off_y = item_off_y + 28 + 10
         screen.blit(cannonSurf, (cannon_off_x, cannon_off_y))
      except:
         pass
      
      if i.HasAny():
         # draw selection box
         sel_off_x = item_off_x + 2 + (28 * self._selection[0])
         sel_off_y = item_off_y + 2 + (28 * self._selection[1])
         pygame.draw.rect(screen, RED, pygame.Rect(sel_off_x, sel_off_y, 24, 24), 1)
   
   
   
class Inventory:
   def __init__(self):
      self._ag = ActiveGame()
   
   def Surf(self, slot):
      i = self._Equipped(slot)
      if i:
         return ImageLib.FromFile(uiImgPath('%s' % i))
      else:
         return None
   
   def HasAny(self):
      i = self
      items = (i.Sabre(), i.Hammer(), i.Drill(), i.Hook(), i.Compass(), i.Cannon(), i.CannonFire(), i.CannonIce(), i.CannonMulti(), i.Shovel())
      for it in items:
         if self.Check(it):
            return True
   
   def Check(self, item):
      return str(self._ag.GetVar(item)) == '1'
   
   def HasSabre(self):
      return self.Check('item_sabre')
   
   def HasHammer(self):
      return self.Check('item_hammer')
   
   def HasDrill(self):
      return self.Checu('item_drill')
   
   def HasHook(self):
      return self.Check('item_hook')
   
   def HasCannon(self):
      return self.Check('item_cannon')
   
   def HasCannonFire(self):
      return self.Check('item_cannon_fire')
   
   def HasCannonIce(self):
      return self.Check('item_cannon_ice')
   
   def HasCannonMulti(self):
      return self.Check('item_cannon_multi')
   
   def HasCompass(self):
      return self.Check('item_compass')
      
   def HasShovel(self):
      return self.Check('item_shovel')
   
   def HasAnyCannon(self):
      return self.HasCannon() or self.HasCannonFire() or self.HasCannonIce() or self.HasCannonMulti()
   
   def Equip(self, button, item):
      slots = ['a', 'b', 'x', 'y']
      
      if not (button == 'a' or button == 'b' or button == 'x' or button == 'y'):
         print('Could not equip to slot "%s"' % str(key))
         return False

      for s in slots:
         if self._Equipped(s) == item:
            self._ag.SetSavedVar('equipped_%s' % s, '')
         if item.startswith('item_cannon') and self._Equipped(s).startswith('item_cannon'):
            self._ag.SetSavedVar('equipped_%s' % s, '')
   
      self._ag.SetSavedVar('equipped_%s' % button, item)
      return True
   
   def EquipA(self, val):
      return self.Equip('a', val)
   def EquipB(self):
      return self.Equip('b', val)
   def EquipX(self, val):
      return self.Equip('x', val)
   def EquipY(self, val):
      return self.Equip('y', val)
   
   def EquippedA(self):
      return self._Equipped('a')
   def EquippedB(self):
      return self._Equipped('b')
   def EquippedX(self):
      return self._Equipped('x')
   def EquippedY(self):
      return self._Equipped('y')
   def _Equipped(self, button):
      i = self._ag.GetVar('equipped_%s' % button) or ''
      return i
   def WhichCannonEquipped(self):
      for l in ['a', 'b', 'x', 'y']:
         item = self._Equipped(l)
         if item and item.startswith('item_cannon'):
            return item
      return False
   
   def Sabre(self):
      return 'item_sabre'
   def Hammer(self):
      return 'item_hammer'
   def Drill(self):
      return 'item_drill'
   def Hook(self):
      return 'item_hook'
   def Cannon(self):
      return 'item_cannon'
   def CannonFire(self):
      return 'item_cannon_fire'
   def CannonIce(self):
      return 'item_cannon_ice'
   def CannonMulti(self):
      return 'item_cannon_multi'
   def Compass(self):
      return 'item_compass'
   def Shovel(self):
      return 'item_shovel'
   
   def Description(self, item):
      table = {}
      table[self.Sabre()] = 'Sabre'
      table[self.Hammer()] = 'Power Hammer'
      table[self.Drill()] = 'Steam Drill'
      table[self.Hook()] = 'Magnetic Gappling Hook'
      table[self.Cannon()] = 'Basic Ammo'
      table[self.CannonFire()] = 'Flame Ammo'
      table[self.CannonIce()] = 'Frost Ammo'
      table[self.CannonMulti()] = 'Multi-shot Ammo'
      table[self.Compass()] = 'Compass'
      table[self.Shovel()] = 'Shovel'
      return table.get(item, '')
   
   def PrintEquipped(self):
      for s in ['a', 'b', 'x', 'y']:
         print('%s => %s' % (s, self._Equipped(s)))

### .\source\Scenes\JoystickConfig.py




class JoystickConfigScene:
	def __init__(self):
		global _inputManager
		self.keys = 'up down left right start X Y A B'.split(' ')
		
		self.text = render_text_size(45, "Push this button", WHITE, MENU_FONT)
		self.error = render_text_size(20, "No Gamepads are plugged in!", WHITE, MENU_FONT)
		self.error_explain = render_text_size(14, "Plug in a USB game pad and relaunch the game", WHITE, MENU_FONT)
		self.texts = [
		'Without a gamepad, use the following keys:',
		' Space - item 1',
		' A - item 2',
		' S - item 3',
		' D - item 4',
		' Enter - inventory screen',
		' Esc - Exit'
		]
		for i in range(len(self.texts)):
			self.texts[i] =  render_text_size(14, self.texts[i], WHITE, MENU_FONT)
			
		self.keyboard = render_text_size(14, "Plug in a USB game pad and relaunch the game", WHITE, MENU_FONT)
		self.is_error = _inputManager.active_joystick == None
		self.render_counter = 1
		self.active_key_index = 0
		self.joystick = _inputManager.active_joystick
		self.done = False
		self.next = self
		
		self.hackery = False
		
		self.images = {
			'up' : get_image('ui/joystick/control_up'),
			'down' : get_image('ui/joystick/control_down'),
			'left' : get_image('ui/joystick/control_left'),
			'right' : get_image('ui/joystick/control_right'),
			'start' : get_image('ui/joystick/control_start'),
			'A' : get_image('ui/joystick/control_a'),
			'B' : get_image('ui/joystick/control_b'),
			'X' : get_image('ui/joystick/control_x'),
			'Y' : get_image('ui/joystick/control_y')
			}
		
		if not self.is_error:
			self.joystick.js.init()
		
	def get_bg_for_key(self, key):
		return self.images[key]
		
	def ProcessInput(self, events):
		if self.is_error or self.done:
			for event in events:
				if event.down:
					if self.is_error:
						self.next = MainMenuScene()
					else:
						if not self.hackery:
							self.hackery = True
						else:
							self.next = MainMenuScene()
							
		
	def Update(self, conter):
		if not self.is_error:
			if self.active_key_index < len(self.keys):
				active_key = self.keys[self.active_key_index]
				if self.joystick.configure_key(active_key):
					self.active_key_index += 1
					if self.active_key_index == len(self.keys):
						self.text = render_text_size(45, "Done!", WHITE, MENU_FONT)
						self.done = True
					
				
		
	def Render(self, screen):
		if self.is_error:
			screen.blit(self.error, (10, 10))
			screen.blit(self.error_explain, (10, 60))
			
			for i in range(len(self.texts)):
				screen.blit(self.texts[i], (10, 100 + i * 20))
			return
		
		image = get_image('ui/joystick/base_control')
		if self.active_key_index < len(self.keys) and (self.render_counter // 15) % 2 == 0:
			image = self.get_bg_for_key(self.keys[self.active_key_index])
		
		screen.blit(image, (0, 0))
		
		screen.blit(self.text, (10, 10))
		
		
		self.render_counter += 1

### .\source\Scenes\MainMenu.py

class MainMenuScene:
   def StartGame(self):
      scene = GameSelectScene(self)
      self.next = scene

   def Setup(self):
      self.next = JoystickConfigScene()
   
   def Credits(self):
      self.next = CreditsScene()
      return
      GameContext().SetActiveGame(1)
      ActiveGame().SetSavedVar('name', 'SUE')
      if os.path.exists('map_test.txt'):
         c = open('map_test.txt', 'rt')
         t = c.read().split('\n')
         c.close()
         map_name = trim(t[0])
         coords = trim(t[1]).split(',')
         scene = GamePlayScene(map_name, int(coords[0]) << 4, int(coords[1]) << 4)
      else:
         scene = GamePlayScene('test_level', 100, 100)
      self.next = scene
      scene.next = scene

   def __init__(self):
      self._fc = 0
      self.next = self
      self._selection = 0
      self._frame = 0
      
      self._gears = []
      i = 1
      while i <= 4:
         self._gears.append(ImageLib.FromFile(uiImgPath('gear%d' % i)))
         i += 1
   
   def ProcessInput(self, events):
      for e in events:
         if e.down:
            if e.Down():
               play_sound("menu2")
               self._selection = (self._selection + 1) % 3
            if e.Up():
               play_sound("menu2")
               self._selection = (self._selection - 1) % 3
            if e.A() or e.B() or e.Start():
               if self._selection == 0:
                  self.StartGame()
               elif self._selection == 1:
                  self.Setup()
               elif self._selection == 2:
                  self.Credits()
            if e.X():
               scene = GameOverScene()
               self.next = scene
               scene.next = scene
            if e.Y():
               scene = CreditsScene()
               self.next = scene
               scene.next = scene

   def Update(self, conter):
      play_music('title')
   
   def Render(self, screen):
      self._fc += 1
      
      frame = self._frame
      if (self._fc % 2 == 0):
         self._frame += 1
         self._frame %= len(self._gears)
      
      title = render_text_size(45, GAME_NAME, WHITE, MENU_FONT)
      start = render_text_size(20, "Start", WHITE, MENU_FONT)
      setup = render_text_size(20, "Setup", WHITE, MENU_FONT) 
      credits = render_text_size(20, "Credits", WHITE, MENU_FONT) 
      art = ImageLib.FromFile(os.path.join('images', 'misc', 'mainmenu-bg.png'))
      
      titleOffset = (int((screen.get_width() - title.get_width()) / 2), 20)
      startOffset = (100, 100)
      setupOffset = (100, 150)
      creditsOffset = (100, 200)
      artOffset = (190, 20)
      
      screen.blit(art, artOffset)
      screen.blit(title, titleOffset)
      screen.blit(start, startOffset)
      screen.blit(setup, setupOffset)
      screen.blit(credits, creditsOffset)
      
      gx = startOffset[0] - 50
      gy = startOffset[1] - 10 + (self._selection * 50)
      g = self._gears[frame]
      screen.blit(g, (gx, gy))


### .\source\Scenes\NameEntry.py

class NameEntryScene:
   def __init__(self):
      self.next = self
      self._selection = [0, 0]
      self._erase = False
      self._done = False
      self._name = 'Lapis'
      self._nameLen = 10
      self._tic = 0
      self._mark = True
   
   def ProcessInput(self, events):
      for e in events:
         if e.down:
            if e.Up():
               play_sound("menu2")
               self._move(0, -1)
            if e.Down():
               play_sound("menu2")
               self._move(0, 1)
            if e.Left():
               self._move(-1, 0)
               play_sound("menu2")
            if e.Right():
               self._move(1, 0)
               play_sound("menu2")
            if (e.A() or e.B() or e.Start()) and self._isCommand():
               if len(self._name) != 0 and self._done:
                  ActiveGame().SetSavedVar('name', self._name)
                  ActiveGame().SaveToFile()
                  scene = GamePlayScene('transport_1', 64, 64)
                  
                  scene.next = scene
                  self.next = scene
                  return
                  
               if self._erase:
                  if len(self._name):
                     self._name = self._name[0:-1]
                     return
            elif e.A() or e.B() or e.Start():
                  if (len(self._name) < self._nameLen):
                     self._name += self._toLetter()
   
   def _toLetter(self):
      s = self._selection
      if s[1] == 4:
         return None
      letters = ['A', 'N', 'a', 'n']
      return chr(ord(letters[s[1]]) + s[0])
   
   def _isCommand(self):
      return self._erase or self._done
   
   def _move(self, x, y):
      s = self._selection
      
      if s[1] == 4:
         if y == 0:
            self._done = not self._done
            self._erase = not self._erase
         else:
            s[1] += y
            s[1] %= 5
            self._done = False
            self._erase = False
      else:
         s[0] += x
         s[0] %= 13
         s[1] += y
         s[1] %= 5
         
         if s[1] == 4:
            if s[0] < 6:
               self._erase = True
            else:
               self._done = True
   
   def Update(self, conter):
      play_music("menuwaitingroom")
      pass
   
   def Render(self, screen):
      self._tic += 1
      sz = 18
      dx = int(sz * 1.3)
      dy = int(sz * 1.5)
      letters = ['A','N','a','n']
      start_off_x = 40
      start_off_y = 80
      
      o_y = start_off_y
      for letter in letters:
         i = 0
         o_x = start_off_x
         if (letters.index(letter) >= 2):
            o_x += int(dx / 6)
         while i < 13:
            c = chr(ord(letter) + i)
            surf = render_text_size(sz, c, WHITE)
            screen.blit(surf, (o_x + int((dx - surf.get_width()) / 2.0), o_y))
            o_x += dx
            i += 1
         o_y += dy
         if (letters.index(letter) == 0):
            o_y += dy / 5
         if (letters.index(letter) == 1):
            o_y += dy / 3
      
      # draw current name
      
      if self._tic % 5 == 0:
         self._mark = not self._mark
      if len(self._name) < self._nameLen and self._mark:
         name = render_text_size(sz, 'Name: %s--' % self._name, WHITE)
      else:
         name = render_text_size(sz, 'Name: %s' % self._name, WHITE)

      screen.blit(name, (start_off_x, start_off_y/2))
      
      # draw Done and Erase
      by = start_off_y + 5 * dy
      
      erase = render_text_size(sz, 'Erase', WHITE)
      ebx = int(screen.get_width() / 4)
      screen.blit(erase, (ebx, by))
      
      done = render_text_size(sz, 'Done', WHITE)
      dbx = 3 * int(screen.get_width() / 5)
      screen.blit(done, (dbx, by))
      
      # draw selection bubble
      s = self._selection
      if s[1] == 4:
         cy = by + erase.get_height() - 5
         if self._erase:
            b = (ebx + 20, cy)
            e = (ebx + erase.get_width(), cy)
            pygame.draw.line(screen, RED, b, e)
         if self._done:
            b = (dbx + 15, cy)
            e = (dbx + done.get_width(), cy)
            pygame.draw.line(screen, RED, b, e)
      else:
         if s[1] < 2:
            x = start_off_x + s[0] * dx + int(dx / 2.0)
            y = start_off_y + s[1] * dy + int(dy / 2.0)
            if s[1] == 1:
               y += int(dy / 7.0)
            pygame.draw.circle(screen, RED, (x, y), int(sz/1.15), 1)
         else:
            x = start_off_x + s[0] * dx + int(dx / 2.0) + int(dx/6)
            y = start_off_y + int(dy/5) + int(dy/3) + s[1] * dy + int(dy / 2.1)
            if s[1] == 1:
               y += int(dy / 7.0)
            pygame.draw.circle(screen, RED, (x, y), int(sz/2), 1)


### .\source\Scenes\SimpleAnimation.py

class SimpleAnimationScene:
	def __init__(self, name):
		self.next = self
		self.name = name
		self.render_counter = 0
		if name == 'flyaway':
			play_sound("flying")
			self.expiration = 60
		elif name == 'flyhome':
			play_sound("flying")
			self.expiration = 60
		else:
			self.expiration = 0

	def ProcessInput(self, events):
		pass

	def Update(self, conter):
		self.expiration -= 1
		if self.expiration <= 0:
			if self.name == 'flyaway':
				self.next = GamePlayScene('escape_pod', 30, 46)
			elif self.name == 'flyhome':
				self.next = GamePlayScene('world_B', 18 *16 + 8, 11*16)
	def Render(self, screen):
		if self.name == 'flyaway':
			x = (4 * self.render_counter) % 384
			y = self.render_counter % 288
			sky = get_image('misc/sky')
			transport = get_image('misc/airship' + str(self.render_counter & 1))
			pod = get_image('misc/escapepod' + str(self.render_counter & 1))
			screen.blit(sky, (x, y))
			screen.blit(sky, (x - 384, y))
			screen.blit(sky, (x, y - 288))
			screen.blit(sky, (x - 384, y - 288))
			
			screen.blit(transport, (60, 10))
			screen.blit(pod, (220 + 3 *  - self.render_counter, 100 + self.render_counter))
		elif self.name == 'flyhome':
			x = (4 * self.render_counter) % 384
			y = self.render_counter % 288
			sky = get_image('misc/sky')
			screen.blit(sky, (x, y))
			screen.blit(sky, (x - 384, y))
			screen.blit(sky, (x, y - 288))
			screen.blit(sky, (x - 384, y - 288))
			pod = get_image('misc/escapepod' + str(self.render_counter & 1))
			screen.blit(pod, (388 + 6 *  - self.render_counter, 120 + self.render_counter))

			
			
		self.render_counter += 1


### .\source\scriptybit.py

# uncomment for testing

class ParserClass:
   def __init__(self):
      self._re_bracket = re.compile('\s*\[(.*?)\]')
   
   # Is the input line a command?
   # Returns:
   #  False if not
   #  The command name if so
   def IsCommand(self, string):
      m = self._re_bracket.match(string)
      if m:
         return m.group(1)
      return False
   
   # Parse a script line into a tuple containing cmd and an array of args
   # Returns:
   #  False if not a command
   #  ($cmd, $args), where args is a list
   def Segment(self, string):
      if not self.IsCommand(string):
         return False
      
      p = self._re_bracket
      m = p.match(string)
      cmd = trim(m.group(1))
      args = []
      
      while True:
         m = p.match(string, m.end())
         if not m:
            break
         args.append(trim(m.group(1)))
      
      return (cmd, args)
   
   # Loads a file into a ScriptIter
   def LoadFile(self, path):
      if not path or not os.path.exists(path):
         print("ERR: %s could not be found" % (str(path)))
         return None
      f = open(path)
      
      scr = []
      for line in f:
         scr.append(trim(line))
      return ScriptIter(scr)
Parser = ParserClass()

# Abstration for the varying sources a script could come from
class ScriptIter:
      # script - an array of strings from any source
      def __init__(self, script):
         self._i = 0
         self._script = script

      # Utility to scan the script for a label.
      # Return:
      #  True - If the label is found, sets index such that Next 
      #         or __next__ will return the label
      #  False - If there is no label
      def FindLabel(self, label):
         idx = 0
         
         tgt = '\s*\[label\]\s*\[%s\]\s*' % (label)
         
         while (idx < len(self._script)):
            if re.match(tgt, self._script[idx]):
               self._i = idx
               return True
            idx += 1
         
         return False

      # Return the next line and step through the script
      def Next(self):
         return self.__next__()

      # Reset the script to the top
      def Reset(self):
         self._i = 0

      def __iter__(self):
         return self
      
      def next(self):
         return self.__next__()

      def __next__(self):
         idx = self._i
         if (idx >= len(self._script)):
            raise StopIteration
         
         self._i += 1
         return self._script[idx]


# encapsulates information about a script command
class Command:
   # cmd ------ text name
   # fn ------- a function pointer
   # argCount - how many args to expect, basic runtime validation
   def __init__(self, cmd, fn, argCount):
      self._name = cmd
      self._fnPtr = fn
      self._argCount = argCount
   
   def Name(self):
      return self._name
   
   def FnPtr(self):
      return self._fnPtr
   
   def ArgCount(self):
      return self._argCount

   # Call this function
   # args - a list of the arguments to call the function with
   def Call(self, args):
      if (len(args) != self.ArgCount()):
         print("ArgMismatch Exception...")
         print("    %s expects %d args" % (self.Name(), self.ArgCount()))
         print("    got '%s'" % (str(args)))
         
      return self.FnPtr().__call__(*args)


# associates strings with function pointers
class CommandRegistry:
   def __init__(self):
      self._commands = {}
   
   # Registers a new command with the scripting engine
   # cmd ------ text name
   # fn ------- a function pointer
   # argCount - how many args to expect, basic runtime validation
   def Register(self, cmd, fn, argCount):
      self._commands[cmd] = Command(cmd, fn, argCount)
   
   # Get a ScriptCommand based on it's name or None if we can't find it
   # cmd - the command to retrieve
   def GetCmd(self, cmd):
      if cmd in self._commands:
         return self._commands[cmd]
      else:
         print("Could not find command %s" % (str(cmd)))
         return None

class ScriptEngine:
   def __init__(self):
      pass
   
# declare globals
cr = CommandRegistry()
se = ScriptEngine()

###############################################################################
# Testing code

def do_test0():
   print("test0")

def do_test1(a):
   print("test1 arg: %s" % (str(a)))

def do_test2(a, b):
   print("test2 arg: %s and %s" % (str(a), str(b)))

def testScript():
   c = CommandRegistry()
   c.Register("test0", do_test0, 0)
   c.Register("test1", do_test1, 1)
   c.Register("test2", do_test2, 2)
   
   args = []
   c.GetCmd("test0").Call(args)
   args.append(1)
   c.GetCmd("test1").Call(args)
   args.append(3)
   c.GetCmd("test2").Call(args)

AddTest('testScript', testScript)

def testScriptIter():
   scr = ['aoeu1', 'aoeu2', 'aoeu3', '[label]  [testing]    ', 'aoeu4', 'aoeu5', 'aoeu6', 'aoeu7']
   si = ScriptIter(scr)
   
   for a in si:
      print(a)
   
   print("---------------------------")
   
   si.FindLabel('testing')
   print(si.Next())
   print(si.Next())
   print(si.Next())
   print("---------------------------")
   si.FindLabel('aoeuaoeu')
   
   for a in si:
      print(a)

AddTest('testScriptIter', testScriptIter)

def testParser():
   tests = ['[jump][fin]', '[check][script-state][eq][done][script-code-done]', 'test', '[end]']
   for a in tests:
      ret = Parser.Segment(a)
      print(ret)
   
   p = scriptPath('test')
   print('loading: %s' % p)
   si = Parser.LoadFile(p)
   print(si)

AddTest('testParser', testParser)

### .\source\script_commands.py

# warp: implement for player warping
# mapFile --------- file path (relative, from base directory)
# tileId ---------- the id of the tile to warp to (alpha numeric)
# transitionStyle - how the warp is displayed, will be one of WARP_X from constants.py
def do_warp(mapFile, tileId, transitionStyle = WARP_INSTANT):
   game_scene = ActiveGame().GetActiveGameScene()
   game_scene.next = TransitionScene(game_scene, mapFile, tileId, transitionStyle)

cr.Register("warp", do_warp, 3)


# dialog: transitions to a dialogue scene using the indicated file as script
# scriptFile - file path (relative, from base directory)
def do_dialog(scriptFile):
   pass
   
cr.Register("dialog", do_dialog, 1)

# remove tile: removes a tile from the map
# tileId ------ id indicating which position and layer
# detailLayer - which detail layer to remove it from
def do_removeTile(tileId, detailLayer):
	return do_setTile(tileId, detailLayer, '')
   
cr.Register('remove tile', do_removeTile, 2)

# set tile: updates the tile at a location
# posId ------- id indicating which position and layer
# detailLayer - which detail layer we're acting in
# tileTypeId -- ID of the tile to add (can be referenced from the tiles.txt file in /data/)
def do_setTile(posId, detailLayer, tileTypeId):
	game_scene = ActiveGame().GetActiveGameScene()
	detailLayer = detailLayer.replace(' ', '').lower()
	if game_scene != None:
		layers = game_scene.level.layers
		ids = game_scene.level.ids
		id = ids.get(posId)
		if id == None:
			print("ERROR: " + posId + ' is not a valid tile ID on this map')
			return
		layers[id.layer].tiles[id.x][id.y].SetTile(detailLayer, tileTypeId)
	return True
cr.Register('set tile', do_setTile, 3)

def do_sound(file):
   play_sound(file)

def do_music(file, loop = False):
   play_music(file)

def do_toggle_mirror(mirror_name):
	current = ActiveGame().GetVar('mirror_state_' + mirror_name)
	if current == 'mirror1': current = 'mirror2'
	elif current == 'mirror2': current = 'mirror3'
	elif current == 'mirror3': current = 'mirror4'
	else: current = 'mirror1'
	play_sound('mirrorrotate')
	ActiveGame().SetSavedVar('mirror_state_' + mirror_name, current)
	#print 'toggle ', mirror_name

def do_switch_scene(newScene):
	game_scene = ActiveGame().GetActiveGameScene()
	if newScene == 'flyaway':
		game_scene.next = SimpleAnimationScene('flyaway')
	if newScene == 'flyhome':
		game_scene.next = SimpleAnimationScene('flyhome')
   
   
def do_cutscene(cutscene):
	game_scene = ActiveGame().GetActiveGameScene()
	game_scene.player.walking = False
	game_scene.cutscene = get_cutscene(cutscene)
	return True

def do_getkey(color):
	game_scene = ActiveGame().GetActiveGameScene()
	dungeon = game_scene.level.dungeon
	GetKeyRegistry().AddKey(dungeon, color)
	return True
	
def do_save_point():
	game_scene = ActiveGame().GetActiveGameScene()
	do_cutscene('save_point_routine')
	return True

def do_sign_display(text):
	lines = trim(str(text)).split('$')
	game_scene = ActiveGame().GetActiveGameScene()
	ds = DialogScene(
		Dialog(ScriptIter(['[profile][]'] + lines + ['[pause]','','[end]'])),
		game_scene)
	game_scene.next = ds
	game_scene.player.walking = False
	return True
	
def do_buy(item):
	price = 0
	if item == 'life':
		price = 5
		var = '_temp'
	if item == 'compass':
		price = 15
		var = 'item_compass'
	elif item == 'shovel':
		price = 20
		var = 'item_shovel'
	elif item == 'fire':
		price = 30
		var = 'item_cannon_fire'
	elif item == 'ice':
		price = 30
		var = 'item_cannon_ice'
	elif item == 'multi':
		price = 50
		var = 'item_cannon_multi'
	elif item == 'armor':
		price = 50
		var = 'has_armor'
	
	if price == 0:
		return True
	
	if has_money(price):
		ActiveGame().SetTempVar('transaction_failed', 0)
		ActiveGame().SetSavedVar(var, 1)
		print('set %s to 1' % str(var))
		modify_money(-price)
		if item == 'life':
			heal_damage()
			heal_damage()
			heal_damage()
			
	else:
		print('failure!')
		ActiveGame().SetTempVar('transaction_failed', '1')
	
	return True
	

### .\source\aaaa.py statics



### .\source\aScripted.py statics



### .\source\Dialog.py statics



### .\source\GameContext.py statics



_gameContext = GameContext()

def ActiveGame():
	global _gameContext
	return _gameContext.GetActiveGame()

def GameContext():
	global _gameContext
	return _gameContext

### .\source\GamePlay\CompassRenderer.py statics



### .\source\GamePlay\CutScene.py statics



_cutSceneStore = { }


_play_once = {
	'interrogation' : False,
	'at_water_temple' : False,
	
}

def get_cutscene(name):
	global _cutSceneStore
	script = _cutSceneStore.get(name)
	if script == None:
		c = open('data' + os.sep + 'cutscenes' + os.sep + name + '.txt', 'rt')
		t = c.read()
		c.close()
		script = trim(t)
		_cutSceneStore[name] = script
		
	return CutScene(script, name)
	
	
def get_cutscene_for_map(map_name):
	global _play_once
	cs = None
	cs_name = None
	if map_name == 'transport_1':
		cs_name = 'interrogation'
	elif map_name == 'escape_pod':
		cs_name = 'to_water_temple'
	elif map_name == 'world_A':
		cs_name = 'at_water_temple'
	
	if cs_name != None:
		cs = get_cutscene(cs_name)
	played_already = _play_once.get(cs_name)
	if played_already != None:
		key = 'cut_scene_play_once_' + cs_name
		if ActiveGame().GetVar(key) == 1:
			cs = None
		else:
			ActiveGame().SetSavedVar(key, 1)
			
	return cs
	

### .\source\GamePlay\DeathCircle.py statics



### .\source\GamePlay\GamePlayScene.py statics


_defaultMirror = {
	'A' : 'mirror1',
	'B' : 'mirror2',
	'C' : 'mirror3',
	'D' : 'mirror4',
	'E' : 'mirror1',
	'F' : 'mirror2',
	'G' : 'mirror3',
	'H' : 'mirror4',
	'I' : 'mirror1',
	'J' : 'mirror2',
	'K' : 'mirror3',
	'L' : 'mirror4',
	'M' : 'mirror1',
	'N' : 'mirror3'
}

### .\source\GamePlay\IdMarker.py statics



### .\source\GamePlay\KeyRegistry.py statics



_key_registry = KeyRegistry()
def GetKeyRegistry():
	global _key_registry
	return _key_registry

### .\source\GamePlay\KeysRenderer.py statics



### .\source\GamePlay\Layer.py statics



### .\source\GamePlay\Level.py statics



### .\source\GamePlay\LevelStore.py statics



### .\source\GamePlay\OverlayRenderer.py statics



### .\source\GamePlay\Sprites\Enemy.py statics



### .\source\GamePlay\Sprites\Goody.py statics



### .\source\GamePlay\Sprites\NPC.py statics



### .\source\GamePlay\Sprites\Player.py statics



### .\source\GamePlay\Sprites\Projectile.py statics



### .\source\GamePlay\Sprites\SpriteFactory.py statics



def create_sprite(name, id=None):
	if ' ' + name + ' ' in ' blob death eyeball skeleton death snake mechanicalman ':
		sprite = Enemy(name, id)
	else:
		sprite = NPC(name, id)
	return sprite

### .\source\GamePlay\Tile.py statics



### .\source\GamePlay\TileStore.py statics



_tileStore = None
def get_tile_store():
	global _tileStore
	if _tileStore == None:
		_tileStore = TileStore()
	return _tileStore
	
	

### .\source\GamePlay\TileTemplate.py statics



### .\source\GamePlay\TransitionScene.py statics


_temp_screen_for_transitions = None
_temp_screens = { }
def GetTempScreen(real_screen):
	global _temp_screen_for_transitions
	if _temp_screen_for_transitions == None:
		w = real_screen.get_width()
		h = real_screen.get_height()
		_temp_screen_for_transitions = pygame.Surface((w, h))
	return _temp_screen_for_transitions
def GetTempScreenSize(width, height):
	global _temp_screens
	index = str(width) + '.' + str(height)
	image = _temp_screens.get(index)
	if image == None:
		image = pygame.Surface((width, height))
		_temp_screens[index] = image
	return image

### .\source\ImageLib.py statics


ImageLib = ImageLibraryClass()

### .\source\ImageLibrary.py statics



_imageLibrary = ImageLibrary()
def get_image(path):
	return _imageLibrary.get_image(path)

### .\source\Input\InputManager.py statics



_inputManager = InputManager()
def is_pressed(key):
	return _inputManager.is_pressed[key]

### .\source\JukeBox.py statics


_jukebox = JukeBox()
def play_sound(name):
	global _jukebox
	_jukebox.PlaySound(name)
def play_music(name):
	global _jukebox
	_jukebox.PlaySong(name)
def reduce_volume():
	global _jukebox
	_jukebox.MakeMusicSoft()
def increase_volume(name):
	global _jukebox
	_jukebox.MakeMusicLoud()
	

### .\source\MapScript.py statics



### .\source\Scenes\BaseScene.py statics



### .\source\Scenes\Credits.py statics



### .\source\Scenes\DialogScene.py statics



### .\source\Scenes\GameOver.py statics



### .\source\Scenes\GameSelect.py statics



### .\source\Scenes\Inventory.py statics



### .\source\Scenes\JoystickConfig.py statics



### .\source\Scenes\MainMenu.py statics



### .\source\Scenes\NameEntry.py statics



### .\source\Scenes\SimpleAnimation.py statics



### .\source\scriptybit.py statics



### .\source\script_commands.py statics




TILE_COLUMN_COUNT = 24
TILE_ROW_COUNT = 18

def main():
	
	counter = 0
	global _font
	
	current_screen_mode = 'windowed'
	
	pygame.init()
	_font = pygame.font.Font(TEXT_FONT, 13)	
	
	width = TILE_COLUMN_COUNT * 16
	height = TILE_ROW_COUNT * 16
	
	screen = pygame.display.set_mode((width * 2, height * 2))
	pygame.display.set_icon(pygame.image.load("icon.png"))
		
	virtual_screen = pygame.Surface((width, height))
	
	_inputManager.initializeJoystick()
	
	scene = MainMenuScene()
	
	screen_width = screen.get_width()
	screen_height = screen.get_height()
	
	while scene != None:
			
		begin = time.time()
		
		scene.ProcessInput(_inputManager.get_events())
		
		scene.Update(counter)
		
		virtual_screen.fill((0,0,0))
		
		scene.Render(virtual_screen)
		
		pygame.transform.scale(virtual_screen, (screen_width, screen_height), screen)
		
		scene = scene.next
		
		change_video_mode = _inputManager.VideoModeChange()
		if change_video_mode != None:
			if current_screen_mode == change_video_mode:
				current_screen_mode = 'windowed'
				screen = pygame.display.set_mode((width * 2, height * 2))
				screen_width = screen.get_width()
				screen_height = screen.get_height()
			else:
				current_screen_mode = change_video_mode
			
				if change_video_mode == 'wide':
					current_screen_mode = 'wide'
					screen = pygame.display.set_mode((1400, 900), FULLSCREEN)
					screen_width = screen.get_width()
					screen_height = screen.get_height()
				else:
					screen = pygame.display.set_mode((width * 2, height * 2), FULLSCREEN)
					screen_width = screen.get_width()
					screen_height = screen.get_height()
		
		if _inputManager.escape_attempted:
			scene = None
		
		counter += 1
		
		pygame.display.flip()
		
		end = time.time()
		
		duration = end - begin
		difference = 1 / 30.0 - duration
		if difference > 0:
			time.sleep(difference)
		else:
			#print("Framerate dropping! (" + str(counter) + ")")
			pass

if testCode:
   RunTests()
else:
   main()